# boto3 implementation of
# https://gist.github.com/shadiakiki1986/f6e676d1ab5800fcf7899b6a392ab821
# Docs
# https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudtrail.html#CloudTrail.Client.get_paginator
#
# Requirements: pip3 install boto3 tqdm pandas
# Run: python3 t2.py
#
# Edit 2019-09-13: copied file pull_cloudtrail_lookupEvents.py from git-remote-aws into isitfit so as to avoid confusion in download statistics
#----------------------------------------

# imports
import datetime as dt
from dateutil.relativedelta import relativedelta
import boto3
import json

from isitfit.utils import logger


#------------------------------
# utility to serialize date
#def json_serial(obj):
#    """JSON serializer for objects not serializable by default json code"""
#
#    if isinstance(obj, (dt.datetime, dt.date)):
#        return obj.isoformat()
#    raise TypeError ("Type %s not serializable" % type(obj))


#----------------------------------------
# iterate

# use jmespath like awscli
# https://stackoverflow.com/a/57018780/4126114
# Example
#   >>> mydata
#   {'foo': {'bar': [{'name': 'one'}, {'name': 'two'}]}}
#   >>> jmespath.search('foo.bar[?name==`one`]', mydata)
#   [{'name': 'one'}]
# import jmespath

#----------------------------------------
class EventIterator:
    eventName = None
  
    # get paginator
    def iterate_page(self):
        """
        eventName - eg 'ModifyInstanceAttribute'
        """
        if self.eventName is None:
          raise Exception("Derived class should set class member eventName")

        # arguments to lookup-events command
        # From docs: "Currently the list can contain only one item"
        LookupAttributes=[
        #    {'AttributeKey': 'EventSource', 'AttributeValue': 'ec2.amazonaws.com'},
            {'AttributeKey': 'EventName', 'AttributeValue': self.eventName},
        ]

        # go back x time
        # https://stackoverflow.com/a/38795526/4126114
        # StartTime=dt.datetime.now() - relativedelta(years=1)
        # StartTime=dt.datetime.now() - relativedelta(days=90)
        PaginationConfig={
          'MaxResults': 3000
        }

        # edit 2019-11-20 instead of defining this client in Gra... and passing it through several layers,
        # just define it here
        # Note 2019-12-09 Cloudtrail can return a max of 90 days
        # In this class, the start/end dates are not specified so as to fetch the whole 90 days and cache them
        # Not very efficient, but works ATM. This is not a per EC2/Redshift call
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudtrail.html#CloudTrail.Client.lookup_events
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudtrail.html#CloudTrail.Paginator.LookupEvents
        client = boto3.client('cloudtrail')
        self.region_name = client.meta.region_name
        cp = client.get_paginator(operation_name="lookup_events")
        iterator = cp.paginate(
          LookupAttributes=LookupAttributes, 
          #StartTime=StartTime, 
          PaginationConfig=PaginationConfig
        )
        return iterator


    def iterate_event(self):
      iter_wrap = self.iterate_page()
      # Update 2019-11-22 moved this tqdm to the region level since it's already super fast per event
      #iter_wrap = tqdm(iter_wrap, desc="Cloudtrail events for %s/%s"%(self.region_name, self.eventName))
      for response in iter_wrap:
        #with open('t2.json','w') as fh:
        #  json.dump(response, fh, default=json_serial)

        # print(response.keys())
        for event in response['Events']:
          result = self._handleEvent(event)
          if result is None: continue
          yield result


    def _handleEvent(self, event):
        # raise Exception("Implement by derived classes")
        return event


class RedshiftCreate(EventIterator):
    eventName = "CreateCluster"
 
    def _handleEvent(self, event):
          # logger.debug("Cloudtrail event: %s"%json.dumps(event, default=json_serial))

          if 'Resources' not in event:
            logger.debug("No 'Resources' key in event. Skipping")
            return None # ignore this situation
        
          instanceId = [x for x in event['Resources'] if x['ResourceType']=='AWS::Redshift::Cluster']
          if len(instanceId)==0:
            logger.debug("No AWS redshift clusters in event. Skipping")
            return None # ignore this situation

          # proceed
          instanceId = instanceId[0]

          if 'ResourceName' not in instanceId:
            logger.debug("No ResourceName key in event. Skipping")
            return None # ignore this situation
          
          # proceed
          instanceId = instanceId['ResourceName']

          if 'CloudTrailEvent' not in event:
            logger.debug("No CloudTrailEvent key in event. Skipping")
            return None # ignore this situation

          ce_dict = json.loads(event['CloudTrailEvent'])

          import jmespath
          nodeType = jmespath.search('requestParameters.nodeType', ce_dict)
          numberOfNodes = jmespath.search('requestParameters.numberOfNodes', ce_dict)
          if numberOfNodes is None:
            numberOfNodes = jmespath.search('responseElements.numberOfNodes', ce_dict)

          if nodeType is None:
            logger.debug("No nodeType key in event['CloudTrailEvent']['requestParameters']. Skipping")
            return None # ignore this situation

          if numberOfNodes is None:
            logger.debug("No numberOfNodes key in event['CloudTrailEvent']['requestParameters']. Skipping")
            return None # ignore this situation

          if 'EventTime' not in event:
            logger.debug("No EventTime key in event. Skipping")
            return None # ignore this situation

          ts_obj = event['EventTime']
          # ts_obj = dt.datetime.utcfromtimestamp(ts_int)
          # ts_str = ts_obj.strftime('%Y-%m-%d %H:%M:%S')

          result = {
            'ServiceName': 'Redshift', # bugfix: was using Ec2 instead of Redshift
            'EventName': self.eventName,
            'EventTime': ts_obj,  # ts_str,
            'ResourceName': instanceId,
            'ResourceSize1': nodeType,
            'ResourceSize2': numberOfNodes,
          }

          return result

      
      
class Ec2Run(EventIterator):
    eventName = "RunInstances"
 
    def _handleEvent(self, event):
          # logger.debug("Cloudtrail event: %s"%json.dumps(event, default=json_serial))

          if 'Resources' not in event:
            logger.debug("No 'Resources' key in event. Skipping")
            return None # ignore this situation
        
          instanceId = [x for x in event['Resources'] if x['ResourceType']=='AWS::EC2::Instance']
          if len(instanceId)==0:
            logger.debug("No AWS EC2 instances in event. Skipping")
            return None # ignore this situation

          # proceed
          instanceId = instanceId[0]

          if 'ResourceName' not in instanceId:
            logger.debug("No ResourceName key in event. Skipping")
            return None # ignore this situation
          
          # proceed
          instanceId = instanceId['ResourceName']

          if 'CloudTrailEvent' not in event:
            logger.debug("No CloudTrailEvent key in event. Skipping")
            return None # ignore this situation

          ce_dict = json.loads(event['CloudTrailEvent'])

          if 'requestParameters' not in ce_dict:
            logger.debug("No requestParameters key in event['CloudTrailEvent']. Skipping")
            return None # ignore this situation

          if 'instanceType' not in ce_dict['requestParameters']:
            logger.debug("No instanceType key in event['CloudTrailEvent']['requestParameters']. Skipping")
            return None # ignore this situation

          newType = ce_dict['requestParameters']['instanceType']

          if 'EventTime' not in event:
            logger.debug("No EventTime key in event. Skipping")
            return None # ignore this situation

          ts_obj = event['EventTime']
          # ts_obj = dt.datetime.utcfromtimestamp(ts_int)
          # ts_str = ts_obj.strftime('%Y-%m-%d %H:%M:%S')

          result = {
            'ServiceName': 'EC2',
            'EventName': self.eventName,
            'EventTime': ts_obj,  # ts_str,
            'ResourceName': instanceId,
            'ResourceSize1': newType,
            'ResourceSize2': None
          }

          return result
          
          
class Ec2Modify(EventIterator):
    eventName = "ModifyInstanceAttribute"
 
    def _handleEvent(self, event):
          if 'CloudTrailEvent' not in event:
            logger.debug("No CloudTrailEvent key in event. Skipping")
            return None # ignore this situation

          ce_dict = json.loads(event['CloudTrailEvent'])

          if 'requestParameters' not in ce_dict:
            logger.debug("No requestParameters key in event['CloudTrailEvent']. Skipping")
            return None # ignore this situation

          rp_dict = ce_dict['requestParameters']
          newType = None

          #newType = jmespath.search('instanceType', rp_dict)
          #if newType is None:
          #  newType = jmespath.search('attributeName==`instanceType`', rp_dict)

          if 'instanceType' in rp_dict:
            # logging.error(json.dumps(rp_dict))
            newType = rp_dict['instanceType']['value']

          if 'attribute' in rp_dict:
            if rp_dict['attribute']=='instanceType':
              newType = rp_dict['value']

          if newType is None:
            return None

          ts_obj = event['EventTime']
          # ts_obj = dt.datetime.utcfromtimestamp(ts_int)
          # ts_str = ts_obj.strftime('%Y-%m-%d %H:%M:%S')

          if 'instanceId' not in rp_dict:
            logger.debug("No instanceId key in requestParameters. Skipping")
            return None # ignore this situation

          result = {
            'ServiceName': 'EC2',
            'EventName': self.eventName,
            'EventTime': ts_obj, # ts_str,
            'ResourceName': rp_dict['instanceId'],
            'ResourceSize1': newType,
            'ResourceSize2': None
          }

          return result


class RedshiftResize(EventIterator):
    eventName = "ResizeCluster"
 
    def _handleEvent(self, event):
          if 'CloudTrailEvent' not in event:
            logger.debug("No CloudTrailEvent key in event. Skipping")
            return None # ignore this situation

          ce_dict = json.loads(event['CloudTrailEvent'])

          if 'requestParameters' not in ce_dict:
            logger.debug("No requestParameters key in event['CloudTrailEvent']. Skipping")
            return None # ignore this situation

          rp_dict = ce_dict['requestParameters']

          import jmespath
          nodeType = jmespath.search('instanceType', rp_dict)
          numberOfNodes = jmespath.search('numberOfNodes', rp_dict)

          ts_obj = event['EventTime']
          # ts_obj = dt.datetime.utcfromtimestamp(ts_int)
          # ts_str = ts_obj.strftime('%Y-%m-%d %H:%M:%S')

          result = {
            'ServiceName': 'Redshift',
            'ResourceName': rp_dict['clusterIdentifier'],
            'EventTime': ts_obj, # ts_str,

            'EventName': self.eventName,
            'ResourceSize1': nodeType,
            'ResourceSize2': numberOfNodes,

          }

          return result


import pandas as pd

class EventAggregatorOneRegion:
    def get(self):
        from termcolor import colored
        import botocore
        import sys

        def run_iterator(man2_i):
          try:
            r_i = list(man2_i.iterate_event())
          except botocore.exceptions.ClientError as e:
            # display error message without the frightening traceback
            from isitfit.cli.click_descendents import IsitfitCliError
            raise IsitfitCliError(str(e))

          return r_i

        man2_ec2run = Ec2Run()
        r_ec2run = run_iterator(man2_ec2run)
        
        man2_ec2mod = Ec2Modify()
        r_ec2mod = run_iterator(man2_ec2mod)
       
        man2_rscre = RedshiftCreate()
        r_rscre = run_iterator(man2_rscre)

        man2_rsmod = RedshiftResize()
        r_rsmod = run_iterator(man2_rsmod)

        # split on instance ID and gather
        r_all = r_ec2run + r_ec2mod + r_rscre + r_rsmod
        # logging.error(r_all)
        df = pd.DataFrame(r_all)

        if df.shape[0]==0:
          # early return
          return df

        df = df.set_index(["ServiceName", "ResourceName", "EventTime"]).sort_index()
        
        return df



class EventAggregatorAllRegions(EventAggregatorOneRegion):
    def __init__(self, region_include, tqdmman):
      self.region_include = region_include
      self.tqdmman = tqdmman

    def get(self):
        # get cloudtrail ec2 type changes for all instances
        logger.debug("Downloading cloudtrail data (from %i regions)"%len(self.region_include))
        df_2 = []
        import boto3

        # add some spaces for aligning the progress bars
        desc="Cloudtrail events in all regions"
        desc = "%-50s"%desc

        iter_wrap = self.region_include
        iter_wrap = self.tqdmman(iter_wrap, desc=desc, total=len(self.region_include))
        for region_name in iter_wrap:
          boto3.setup_default_session(region_name = region_name)
          df_1 = super().get()
          df_1['Region'] = region_name # bugfix, field name was "region" (lower-case)
          df_2.append(df_1.reset_index())

        # concatenate
        df_3 = pd.concat(df_2, axis=0, sort=False)

        # check if empty
        if df_3.shape[0]==0:
          return df_3

        # sort again
        df_3 = df_3.set_index(["Region", "ServiceName", "ResourceName", "EventTime"]).sort_index()

        return df_3




class EventAggregatorCached(EventAggregatorAllRegions):
    cache_key = "cloudtrail_ec2type._fetch"

    def __init__(self, region_include, tqdmman, cache_man):
        super().__init__(region_include, tqdmman)
        self.cache_man = cache_man


    def get(self):
        # get cloudtrail ec2 type changes for all instances

        # if not configured, just return
        if self.cache_man is None:
          df_fresh = super().get()
          return df_fresh

        # check cache first
        if self.cache_man.isReady():
          df_cache = self.cache_man.get(self.cache_key)
          if df_cache is not None:
            logger.debug("Found cloudtrail data in redis cache")
            return df_cache

        # if no cache, then download
        df_fresh = super().get()

        # if caching enabled, store it for later fetching
        # https://stackoverflow.com/a/57986261/4126114
        if self.cache_man.isReady():
          self.cache_man.set(self.cache_key, df_fresh)

        # done
        return df_fresh





def dict2service(ec2_dict):
        if 'InstanceId' in ec2_dict: return 'EC2'
        if 'ClusterIdentifier' in ec2_dict: return 'Redshift'
        import json
        raise Exception("Unknown service found in %s"%json.dumps(ec2_dict))



class EventAggregatorPostprocessed(EventAggregatorCached):
    def __init__(self, region_include, tqdmman, cache_man, EndTime):
        super().__init__(region_include, tqdmman, cache_man)
        self.EndTime = EndTime


    def get(self, ec2_instances, n_ec2):
        self.df_cloudtrail = super().get()

        # first pass to append ec2 types to cloudtrail based on "now"
        self.df_cloudtrail = self.df_cloudtrail.reset_index()

        # add some spaces for aligning the progress bars
        desc = "Pass 1/2 Cloudtrail history (EC2, Redshift)"
        desc = "%-50s"%desc

        # Edit 2019-11-12 use initial=0 otherwise if "=1" used then the tqdm output would be "101it" at conclusion, i.e.
        # First pass through EC2 instances: 101it [00:05,  5.19it/s]
        t_iter = ec2_instances
        t_iter = self.tqdmman(t_iter, total=n_ec2, desc=desc, initial=0)
        for ec2_dict, ec2_id, ec2_launchtime, ec2_obj in t_iter:
            self._appendNow(ec2_dict, ec2_id)

        # if still no data, just return
        if self.df_cloudtrail.shape[0]==0:
          return self.df_cloudtrail

        # set index again, and sort decreasing this time (not like git-remote-aws default)
        # The descending sort is very important for the mergeTimeseries... function
        self.df_cloudtrail = self.df_cloudtrail.set_index(["Region", "ServiceName", "ResourceName", "EventTime"]).sort_index(ascending=False)

        # done
        return self.df_cloudtrail


    def _appendNow(self, ec2_dict, ec2_id):
        # artificially append an entry for "now" with the current type
        # This is useful for instance who have no entries in the cloudtrail
        # so that their type still shows up on merge

        ec2_dict['ServiceName'] = dict2service(ec2_dict)

        size1_key = 'NodeType' if ec2_dict['ServiceName']=='Redshift' else 'InstanceType'
        size2_val = ec2_dict['NumberOfNodes'] if ec2_dict['ServiceName']=='Redshift' else None

        df_new = pd.DataFrame([
              {
                'Region': ec2_dict['Region'],
                'ServiceName': ec2_dict['ServiceName'],
                'ResourceName': ec2_id,
                'EventTime': self.EndTime,
                'ResourceSize1': ec2_dict[size1_key],
                'ResourceSize2': size2_val
              }
            ])

        self.df_cloudtrail = pd.concat([self.df_cloudtrail, df_new], sort=True)




"""
Unit tests for python exception handling.
Purpose is to understand how this works exactly
"""


import pytest


def test_finally():
  class DoIt:
    x=1
    def doit(self):
      try:
        raise Exception
      except ValueError:
        self.x=2
      finally:
        self.x=3

  y=DoIt()
  with pytest.raises(Exception):
    y.doit()
    assert False # should not be reached

  # after exception
  assert y.x==3


def test_multiExcept():
  class E1(ValueError): pass
  class E2(ValueError): pass

  def doit(et):
    x=1
    try:
      raise et
    except E1:
      x=2
    except E2:
      x=3
    except:
      raise

    return x

  assert doit(E1())==2
  assert doit(E2())==3

  with pytest.raises(Exception):
    doit(Exception())



from isitfit.utils import logger


from isitfit.utils import mergeSeriesOnTimestampRange


class Ec2Common:
    def _handle_ec2obj(self, context_ec2):
        # parse out
        ec2_obj = context_ec2['ec2_obj']

        # logger.debug("%s, %s"%(ec2_obj.instance_id, ec2_obj.instance_type))

        # pandas series of CPU utilization, daily max, for 90 days
        df_metrics = context_ec2['df_metrics']

        # pandas series of number of cpu's available on the machine over time, past 90 days
        df_type_ts1 = context_ec2['df_type_ts1']
        df_type_ts1 = df_type_ts1.rename(columns={'ResourceSize1': 'instanceType'})

        # this is redundant with the implementation in _cloudwatch_metrics_core,
        # and it's here just in case the cached redis version is not a date,
        # but it's not really worth it to make a full refresh of the cache for this
        # if df_metrics.Timestamp.dtype==dt.date:
        try:
          df_metrics['Timestamp'] = df_metrics.Timestamp.dt.date
        except AttributeError:
          pass

        # convert type timeseries to the same timeframes as pcpu and n5mn
        #if ec2_obj.instance_id=='i-069a7808addd143c7':
        ec2_df = mergeSeriesOnTimestampRange(df_metrics, df_type_ts1, ['instanceType'])
        #logger.debug("\nafter merge series on timestamp range")
        #logger.debug(ec2_df.head())

        # merge with type changes (can't use .merge on timestamps, need to use .concat)
        #ec2_df = df_metrics.merge(df_type_ts2, left_on='Timestamp', right_on='EventTime', how='left')
        # ec2_df = pd.concat([df_metrics, df_type_ts2], axis=1)

        # merge with catalog
        ec2_df = ec2_df.merge(context_ec2['df_cat'][['API Name', 'cost_hourly']], left_on='instanceType', right_on='API Name', how='left')
        #logger.debug("\nafter merge with catalog")
        #logger.debug(ec2_df.head())


        # append instance ID and region for clarity
        ec2_df['region'] = ec2_obj.region_name
        ec2_df['instance_id'] = ec2_obj.instance_id

        # and move these columns to the start
        cols_first = ['region', 'instance_id']
        cols_exRegId = [x for x in ec2_df.columns if x not in cols_first]
        ec2_df = ec2_df[cols_first + cols_exRegId]

        # set in context
        context_ec2['ec2_df'] = ec2_df

        # done
        return context_ec2


    def after_all(self, context_all):
        # unpack
        # Update 2019-12-17 ec2_noCloudwatch is deprecated
        # ec2_noCloudwatch, ec2_noCloudtrail = context_all['ec2_noCloudwatch'], context_all['ec2_noCloudtrail']
        ec2_noCloudtrail = context_all['ec2_noCloudtrail']


        # get now + 10 minutes
        # http://stackoverflow.com/questions/6205442/ddg#6205529
        import datetime as dt
        dt_now = dt.datetime.now()
        TRY_IN = 10
        now_plus_10 = dt_now + dt.timedelta(minutes = TRY_IN)
        now_plus_10 = now_plus_10.strftime("%H:%M")

#        if len(ec2_noCloudwatch)>0:
#          n_no_cw = len(ec2_noCloudwatch)
#          has_more_cw = "..." if n_no_cw>5 else ""
#          l_no_cw = ", ".join(ec2_noCloudwatch[:5])
#          logger.info("No cloudwatch data for %i resources: %s%s"%(n_no_cw, l_no_cw, has_more_cw))
#          logger.info("Try again in %i minutes (at %s) to check for new data"%(TRY_IN, now_plus_10))
#          logger.info("")

        if len(ec2_noCloudtrail)>0:
          n_no_ct = len(ec2_noCloudtrail)
          has_more_ct = "..." if n_no_ct>5 else ""
          l_no_ct = ", ".join(ec2_noCloudtrail[:5])
          logger.info("No cloudtrail data for %i resources: %s%s"%(n_no_ct, l_no_ct, has_more_ct))
          logger.info("Try again in %i minutes (at %s) to check for new data"%(TRY_IN, now_plus_10))
          logger.info("")

        return context_all



def tagsContain(f_tn, ec2_obj):
  """
  Similar to isitfit.cost.redshift_common.tagsContain
  """
  if ec2_obj.tags is None:
    return False
  
  for t in ec2_obj.tags:
    for k in ['Key', 'Value']:
      if f_tn in t[k].lower():
        return True
  
  return False



class Ec2TagFilter:
  def __init__(self, filter_tags):
    self.filter_tags = filter_tags

  def per_ec2(self, context_ec2):
    # if filters requested, check that this instance passes

    # set in context for the sake of filtering the taglist in ec2_optimize.Calculator.per_ec2
    context_ec2['filter_tags'] = self.filter_tags

    if self.filter_tags is None:
      # to continue with other listeners
      return context_ec2

    f_tn = self.filter_tags.lower()
    passesFilter = tagsContain(f_tn, context_ec2['ec2_obj'])
    if not passesFilter:
      # break other listeners
      return None

    # otherwise continue
    return context_ec2



from isitfit.cost.metrics_cloudwatch import CloudwatchAssistant, NoCloudwatchException

import pytest

class TestCwAssFunctional:
  def test_tableauDesktopInstance_noData(self):
    import datetime as dt
    dtnow = dt.datetime.now()

    ca = CloudwatchAssistant()
    ca.set_resource("us-west-2")
    assert True # no exception

    iid = 'i-05a61a1d9c3d208b3'
    it = ca.id2iterator(iid, 'AWS/EC2', 'InstanceId')
    assert it is not None

    with pytest.raises(NoCloudwatchException):
      me = ca.iterator2metric(it, iid)


  def test_inexistantId(self):
    import datetime as dt
    dtnow = dt.datetime.now()

    ca = CloudwatchAssistant()
    ca.set_resource("us-west-2")
    assert True # no exception

    iid = 'i-123456'
    it = ca.id2iterator(iid, 'AWS/EC2', 'InstanceId')
    assert it is not None

    with pytest.raises(NoCloudwatchException):
      me = ca.iterator2metric(it, iid)


  def test_newInstance_ndays90(self):
    """
    Run this after creating an instance manually
    (with spot using isitfit/tests/cost/ec2/test_functional.sh)
    and then plugging in the ID here
    """
    import datetime as dt
    dtnow = dt.datetime.now()

    ca = CloudwatchAssistant()
    ca.set_resource("us-west-2")
    assert True # no exception

    iid = 'i-0554591aacf06353a' # <<<<<<<< ID of newly created instance here
    it = ca.id2iterator(iid, 'AWS/EC2', 'InstanceId')
    assert it is not None

    me = ca.iterator2metric(it, iid)
    assert me is not None

    # test with ndays=90
    ca.set_ndays(90)
    st90 = ca.metric2stats(me)
    assert st90 is not None
    # for ndays=90, this is not a reliable assertion
    # assert st90['Datapoints'][0]['Timestamp'].date() == dtnow.date()

    df90 = ca.stats2df(st90, iid, dtnow)
    assert df90 is not None
    assert df90.shape[0] == 1
    # this gets corrected if cloudwatch returns a past timestamp for an instance created today
    assert df90.Timestamp.iloc[0] == dtnow.date()

    # test with ndays=7
    ca.set_ndays(7)
    st07 = ca.metric2stats(me)
    assert st07 is not None
    # for ndays=7, this is always failing
    assert st07['Datapoints'][0]['Timestamp'].date() != dtnow.date()

    df07 = ca.stats2df(st07, iid, dtnow)
    assert df07 is not None
    assert df07.shape[0] == 1
    # stats2df will fix the wrong timestamp
    assert df07.Timestamp.iloc[0] == dtnow.date()



"""
These tests requires firing up a new ec2 instance and installing the datadog agent on it.
Check isitfit/tests/cost/ec2/test_functional.sh for helper commands

Maybe need to use pytest.mark.skip on these tests, or move them to separate folder with some special automatic ec2 instance provisioning
"""

# Update this instance ID when turning on a new instance with a datadog agent running
HOST_ID_VALID = 'i-0a22ca422859d1cd9'

from isitfit.cost.metrics_datadog import DatadogApiWrap, DataQueryError, DataNotFoundForHostInDdg


import pytest

from isitfit.utils import SECONDS_IN_ONE_DAY


@pytest.fixture(scope='session')
def datadog_api():
    import datadog
    datadog.initialize()
    return datadog.api



def get_startEnd():
    # set start/end
    # Update, use time.time() method instead
    #import datetime as dt
    #from datetime import timedelta
    #dt_now = dt.datetime.now()
    #dt_1w  = dt_now - timedelta(days=7)
    #conv2sec = lambda x: time.mktime(x.timetuple())
    #ue_start = conv2sec(start)
    #ue_end   = conv2sec(end)
    import time
    dt_now = int(time.time())
    dt_1w  = dt_now - SECONDS_IN_ONE_DAY*7
    return dt_now, dt_1w


class TestDatadogApiWrapFunctional:
  """
  Tests to help resolve
  https://github.com/autofitcloud/isitfit/issues/10
  """
  def test_hosts_search_ok(self, datadog_api):
    daw = DatadogApiWrap()
    host_id = HOST_ID_VALID
    h_i = daw.hosts_search(host_id)
    #assert h_i['name']==host_id

    # this should match the ec2 type selected when launching this test instance
    assert h_i['cpuCores']==1
    assert int(h_i['memory_total']/1e9)==4


  def test_hosts_search_notFound(self, datadog_api):
    daw = DatadogApiWrap()
    host_id = 'i-1234566'
    from isitfit.utils import HostNotFoundInDdg
    with pytest.raises(HostNotFoundInDdg):
      h_i = daw.hosts_search(host_id)


  def test_metric_query_cpuIdle_ok(self, datadog_api):
    daw = DatadogApiWrap()
    host_id = HOST_ID_VALID

    # set start/end
    dt_now, dt_1w = get_startEnd()

    query = 'system.cpu.idle{host:%s}.rollup(min,%i)'%(host_id, SECONDS_IN_ONE_DAY)
    df = daw.metric_query(host_id, dt_1w, dt_now, query, 'system.cpu.idle', 'some_name')

    assert df.shape[0] > 0
    assert 'some_name' in df.columns


  def test_metric_query_cpuIdle_errQuery(self, datadog_api):
    daw = DatadogApiWrap()
    host_id = HOST_ID_VALID # exists

    # set start/end
    dt_now, dt_1w = get_startEnd()

    query = 'blabla'
    with pytest.raises(DataQueryError):
      df = daw.metric_query(host_id, dt_1w, dt_now, query, 'inexistant', 'somename')



  def test_metric_query_cpuIdle_errNotFound(self, datadog_api):
    daw = DatadogApiWrap()
    host_id = 'i-12345'

    # set start/end
    dt_now, dt_1w = get_startEnd()

    from isitfit.utils import SECONDS_IN_ONE_DAY
    query = 'system.cpu.idle{host:%s}.rollup(min,%i)'%(host_id, SECONDS_IN_ONE_DAY)

    with pytest.raises(DataNotFoundForHostInDdg):
      df = daw.metric_query(host_id, dt_1w, dt_now, query, 'inexistant', 'somename')





from ...cost.metrics_datadog import DatadogManager, DatadogAssistant, DataNotFoundForHostInDdg, HostNotFoundInDdg, DatadogCached
import pandas as pd
import pytest



@pytest.fixture
def datadog_assistant(mocker):
    def factory(series=[], host_list=[]):
      mockreturn = lambda *args, **kwargs: {'series': series, 'status': 'ok'}
      mockee = 'datadog.api.Metric.query'
      mocker.patch(mockee, side_effect=mockreturn)

      mockreturn = lambda *args, **kwargs: {'total_returned': len(host_list), 'host_list': host_list}
      mockee = 'datadog.api.Hosts.search'
      mocker.patch(mockee, side_effect=mockreturn)

      # set start/end
      import datetime as dt
      from datetime import timedelta
      dt_now = dt.datetime.now()
      dt_1w  = dt_now - timedelta(days=7)

      dda = DatadogAssistant(dt_1w, dt_now, host_id='i-123456')
      return dda

    return factory


class TestDatadogAssistant:

  def test_init(self, datadog_assistant):
    dda = datadog_assistant([])
    assert True

  def test_getMetricsCore_noData(self, datadog_assistant):
    dda = datadog_assistant(series=[])
    with pytest.raises(DataNotFoundForHostInDdg):
      df = dda._get_metrics_core(None, 'Average')

  def test_getMetricsCore_ok(self, datadog_assistant):
    dda = datadog_assistant(series=[{'pointlist': [{'ts_int': 1234567, 'Average': 2}]}])
    df = dda._get_metrics_core(None, 'Average')
    assert df.shape[0]==1

  def test_getMeta_noData(self, datadog_assistant):
    dda = datadog_assistant(host_list=[])
    with pytest.raises(HostNotFoundInDdg):
      df = dda._get_meta()

  def test_getMeta_ok(self, datadog_assistant):
    dda = datadog_assistant(host_list=[{'meta': {'gohai': '{"memory": {"total": "10kB"}}', 'cpuCores': 2}, 'name': 'i-123456'}])
    res = dda._get_meta()
    assert res=={'cpuCores': 2, 'memory_total': 10240}

  def test_getMetricsXxxYyy(self, datadog_assistant):
    pointlist = [{'ts_int': 1234567, 'cpu_idle_min': 2.5, 'cpu_idle_avg': 2.5, 'ram_free_min': 10, 'ram_free_avg': 5}]
    dda = datadog_assistant(
      series=[{'pointlist': pointlist}],
      host_list=[{'meta': {'gohai': '{"memory": {"total": "10kB"}}', 'cpuCores': 2}, 'name': 'i-123456'}],
    )

    actual = dda.get_metrics_cpu_max()
    assert actual.shape[0]==1
    assert actual.shape[1]==3 # columns:  cpu_idle_min                   ts_dt  cpu_used_max

    actual = dda.get_metrics_cpu_avg()
    assert actual.shape[0]==1
    assert actual.shape[1]==3

    actual = dda.get_metrics_ram_max()
    assert actual.shape[0]==1
    assert actual.shape[1]==3

    actual = dda.get_metrics_ram_avg()
    assert actual.shape[0]==1
    assert actual.shape[1]==3


@pytest.fixture
def datadog_manager(mocker):
  def factory():
      mockreturn = lambda *args, **kwargs: None
      mockee = 'datadog.initialize'
      mocker.patch(mockee, side_effect=mockreturn)
      ddm = DatadogManager()
      return ddm

  return factory


@pytest.fixture
def ddgenv_set(monkeypatch):
     monkeypatch.setenv('DATADOG_API_KEY', 'abcdef')
     monkeypatch.setenv('DATADOG_APP_KEY', 'abcdef')


@pytest.fixture
def ddgenv_missing(monkeypatch):
     monkeypatch.delenv('DATADOG_API_KEY')
     monkeypatch.delenv('DATADOG_APP_KEY')


class TestDatadogManager:
  def test_init(self, datadog_manager):
      ddm = datadog_manager()
      assert True # no exception

  def test_isConfigured_true(self, datadog_manager, ddgenv_set):
      ddm = datadog_manager()
      assert ddm.is_configured()

  def test_isConfigured_false(self, datadog_manager, ddgenv_missing):
      ddm = datadog_manager()
      assert not ddm.is_configured()


  def test_getMetricsAll(self, datadog_manager, mocker):
      def mymock(mr, me):
        mockreturn = lambda *args, **kwargs: mr
        mockee = 'isitfit.cost.metrics_datadog.DatadogAssistant.%s'%me
        mocker.patch(mockee, side_effect=mockreturn)

      import datetime as dt
      dtnow = dt.datetime.now()

      mockees = [
        ( pd.DataFrame({'ts_int':[1], 'ts_dt':[dtnow], 'cpu_used_max': [3]}),
         'get_metrics_cpu_max'
        ),
        ( pd.DataFrame({'ts_int':[1], 'ts_dt':[dtnow], 'cpu_used_avg': [3]}),
          'get_metrics_cpu_avg'
        ),
        ( pd.DataFrame({'ts_int':[1], 'ts_dt':[dtnow], 'cpu_used_min': [3]}),
         'get_metrics_cpu_min'
        ),
        ( pd.DataFrame({'ts_int':[1], 'ts_dt':[dtnow], 'ram_used_max': [3]}),
          'get_metrics_ram_max'
        ),
        ( pd.DataFrame({'ts_int':[1], 'ts_dt':[dtnow], 'ram_used_avg': [3]}),
          'get_metrics_ram_avg'
        ),
        ( pd.DataFrame({'ts_int':[1], 'ts_dt':[dtnow], 'ram_used_min': [3]}),
          'get_metrics_ram_min'
        ),
        ( pd.DataFrame({'ts_int':[1], 'ts_dt':[dtnow], 'nhours': [3]}),
          'get_metrics_count'
        ),
      ]
      for mr, me in mockees: mymock(mr, me)

      ddm = datadog_manager()
      actual = ddm.get_metrics_all('i-123456')

      assert actual.shape[0]==1
      assert actual.shape[1]==8 # columns: ts_dt  cpu_used_max  cpu_used_avg  cpu_used_min  ram_used_max  ram_used_avg  ram_used_min  nhours



@pytest.fixture
def cache_man(mocker):
    """
    Mocked cache manager
    """
    class MockCacheMan:
      def __init__(self):
        self._map = {}
        self.ready = False

      def isReady(self): return self.ready
      def get(self, key): return self._map.get(key)
      def set(self, key, val): self._map[key] = val

    cache_man = MockCacheMan()
    mocker.spy(cache_man, 'get')
    mocker.spy(cache_man, 'set')
    return cache_man


class TestDatadogCachedGetMetricsDerived:
  def test_notReady_noData(self, mocker, cache_man):
    # mock parent
    # mockreturn = lambda *args, **kwargs: pd.DataFrame()
    def mockreturn(*args, **kwargs): raise DataNotFoundForHostInDdg
    mockee = 'isitfit.cost.metrics_datadog.DatadogManager.get_metrics_all'
    mocker.patch(mockee, side_effect=mockreturn)

    ddc = DatadogCached(cache_man)

    # after first call
    with pytest.raises(DataNotFoundForHostInDdg):
      actual = ddc.get_metrics_derived(None, 'i-123456', None)

    assert cache_man.get.call_count == 0
    assert cache_man.set.call_count == 0
    assert cache_man._map == {}

    # after 2nd call
    with pytest.raises(DataNotFoundForHostInDdg):
      actual = ddc.get_metrics_derived(None, 'i-123456', None)

    assert cache_man.get.call_count == 0
    assert cache_man.set.call_count == 0
    assert cache_man._map == {}


  def test_notReady_yesData(self, mocker, cache_man):
    # mock parent
    mockreturn = lambda *args, **kwargs: pd.DataFrame({'a': [1,2,3]})
    mockee = 'isitfit.cost.metrics_datadog.DatadogManager.get_metrics_all'
    mocker.patch(mockee, side_effect=mockreturn)

    ddc = DatadogCached(cache_man)
    actual = ddc.get_metrics_derived(None, 'i-123456', None)
    assert actual is not None
    assert actual.shape[0]==3

    assert cache_man.get.call_count == 0
    assert cache_man.set.call_count == 0
    assert cache_man._map == {}


  def test_yesReady_noData(self, mocker, cache_man):
    # mark as ready
    cache_man.ready = True

    # mock parent
    # mockreturn = lambda *args, **kwargs: pd.DataFrame()
    def mockreturn(*args, **kwargs): raise DataNotFoundForHostInDdg
    mockee = 'isitfit.cost.metrics_datadog.DatadogManager.get_metrics_all'
    mocker.patch(mockee, side_effect=mockreturn)

    ddc = DatadogCached(cache_man)
    host_id = 'i-123456'

    # after first call
    with pytest.raises(DataNotFoundForHostInDdg):
      actual = ddc.get_metrics_derived(None, host_id, None)

    assert cache_man.get.call_count == 1 # checks cache and doesn't find key
    assert cache_man.set.call_count == 1 # first set key
    assert callable(cache_man._map[ddc.get_key(host_id)])

    # after 2nd call
    with pytest.raises(DataNotFoundForHostInDdg):
      actual = ddc.get_metrics_derived(None, host_id, None)

    assert cache_man.get.call_count == 2 # incremented
    assert cache_man.set.call_count == 1 # no increment
    assert callable(cache_man._map[ddc.get_key(host_id)])


  def test_yesReady_invalidCache(self, mocker, cache_man):
    # enable mocked cache
    cache_man.ready = True

    # set key in cache
    host_id = 'i-123456'
    ddc = DatadogCached(cache_man)
    cache_man._map[ddc.get_key(host_id)] = pd.DataFrame()

    # fetch will raise
    with pytest.raises(Exception):
      actual = ddc.get_metrics_derived(None, host_id, None)


  def test_yesReady_yesData(self, mocker, cache_man):
    # mark as ready
    cache_man.ready = True

    # mock parent
    mockreturn = lambda *args, **kwargs: pd.DataFrame({'a': [1,2,3]})
    mockee = 'isitfit.cost.metrics_datadog.DatadogManager.get_metrics_all'
    uncached_get = mocker.patch(mockee, side_effect=mockreturn)

    ddc = DatadogCached(cache_man)
    host_id = 'i-123456'

    # after first call
    actual = ddc.get_metrics_derived(None, host_id, None)
    assert actual is not None
    assert actual.shape[0]==3
    assert uncached_get.call_count == 1 # calls upstream
    assert cache_man.get.call_count == 1 # 1st check in cache
    assert cache_man.set.call_count == 1 # 1st set in cache

    # after 2nd call
    actual = ddc.get_metrics_derived(None, host_id, None)
    assert actual is not None
    assert actual.shape[0]==3
    assert uncached_get.call_count == 1 # no increment
    assert cache_man.get.call_count == 2 # 2nd check in cache
    assert cache_man.set.call_count == 1 # no increment



"""
pytest tests to figure out a solution for https://github.com/autofitcloud/isitfit/issues/10
"Missing memory data"
isitfit unable to pick up data from datadog for a particular instance, despite the data's availability
"""

import pytest

@pytest.fixture(scope='session')
def datadog_api():
    import datadog
    datadog.initialize()
    return datadog.api


class TestIssue10:
  """
  Tests to help resolve
  https://github.com/autofitcloud/isitfit/issues/10
  """
  datadog_hostname = None

  def test_hosts(self, datadog_api):
    # check total numbers and display message in case of failure later
    h_all = datadog_api.Hosts.totals()
    print("Found {total_up} ec2 hosts that are UP, and {total_active} that are active".format(**h_all))
    assert h_all['total_up'] > 0
    assert h_all['total_active'] > 0

    # search without filter
    # print a list of 5 ec2 IDs in case of error when searching for specific ID below
    MAX_HOSTS = 5 # limit just in case
    h_search_count = datadog_api.Hosts.search(count=MAX_HOSTS)
    print("first %i IDs found (aws_id, name, host_name): "%MAX_HOSTS, [tuple(hi.get(k) for k in ['aws_id', 'name', 'host_name']) for hi in h_search_count['host_list']])
    print("datadog_api.Hosts.search(count=%i) full output:"%MAX_HOSTS)
    for i, hi in enumerate(h_search_count['host_list']): print("%i: "%i, hi)
    for i, hi in enumerate(h_search_count['host_list']): assert 'aws_id' in hi
    assert len(h_search_count['host_list']) > 0
    assert h_search_count['total_returned'] > 0

    # now search for provided ID
    h_all = datadog_api.Hosts.search(filter='host:%s'%self.datadog_hostname)
    assert len(h_all['host_list']) > 0
    assert h_all['total_returned'] == 1
    # alternatively can check host_name
    assert h_all['host_list'][0]['name']==self.datadog_hostname

    # you shall not pass
    assert False


  def test_metric_query_cpuIdle(self, datadog_api):
    # set start/end
    import datetime as dt
    from datetime import timedelta
    dt_now = dt.datetime.now()
    dt_1w  = dt_now - timedelta(days=7)

    # convert to seconds since unix epoch
    # https://stackoverflow.com/a/6999787/4126114
    import time
    conv2sec = lambda x: time.mktime(x.timetuple())
    ue_now = conv2sec(dt_now)
    ue_1w  = conv2sec(dt_1w)

    # build query
    from isitfit.utils import SECONDS_IN_ONE_DAY
    query = 'system.cpu.idle{host:%s}.rollup(min,%i)'%(self.datadog_hostname, SECONDS_IN_ONE_DAY)

    # query datadog
    # https://docs.datadoghq.com/api/?lang=python#query-timeseries-points
    m = datadog_api.Metric.query(start=ue_1w, end=ue_now, query=query)

    if 'errors' in m:
      print(m)
      raise Exception(m['errors'])

    if m['status'] != 'ok':
      raise Exception(m['status'])

    assert len(m['series'])>0


import click
from isitfit.cli.click_descendents import IsitfitCommand
@click.command(help='Tests to debug github issue #10', cls=IsitfitCommand, hidden=True)
@click.argument('datadog_hostname') #, help='Instance ID on which to run this test')
def issue10(datadog_hostname):
  """
  Click command that runs test through pytest
  """
  TestIssue10.datadog_hostname = datadog_hostname

  # https://docs.pytest.org/en/latest/usage.html#calling-pytest-from-python-code
  exit_code = pytest.main([__file__, '--verbose'])
  conclude_msg = ""
  conclude_color = ""
  if exit_code != 0:
    conclude_msg = "isitfit: Tests for issue #10 failed"
    conclude_color = "red"
  else:
    conclude_msg = "isitfit: Tests for issue #10 passed"
    conclude_color = "green"

  import click
  click.secho(conclude_msg, fg=conclude_color)
  click.secho("More details at https://github.com/autofitcloud/isitfit/issues/10")



from isitfit.cost.redshift_analyze import ReporterAnalyze
from isitfit.cost.base_reporter import ReporterBase
from isitfit.cost.redshift_optimize import ReporterOptimize


class TestReporterBase:
  def test_init(self):
    rb = ReporterBase()
    assert True



class TestReporterAnalyze:
  def test_postprocess(self):
    import datetime as dt
    dt_now = dt.datetime.utcnow()

    class MockMm:
      StartTime = dt_now
      EndTime = dt_now

    class MockAnalyzer:
      class MockIter:
        pass

      cwau_percent = 10
      rp_iter = MockIter
      regions_n = 1
      cost_billed = 1
      cost_used = 1

    rb = ReporterAnalyze()
    rb.postprocess({'analyzer': MockAnalyzer, 'mainManager': MockMm, 'n_ec2_total': 1, 'n_rc_analysed': 0})
    assert rb.table is not None



class TestReporterOptimize:
  def test_postprocess(self):
    import pandas as pd
    class MockAnalyzer:
      analyze_df = pd.DataFrame([{'CpuMaxMax': 1, 'CpuMinMin': 1}])

    rb = ReporterOptimize()
    rb.postprocess({'analyzer': MockAnalyzer})
    assert True # no exception


  def test_display(self, mocker):
    mockee = 'isitfit.utils.display_df'
    mocker.patch(mockee, autospec=True)

    import pandas as pd
    class MockAnalyzer:
      analyze_df = pd.DataFrame([{'CpuMaxMax': 1, 'CpuMinMin': 1}])

    rb = ReporterOptimize()
    rb.csv_fn_final = 'bla.csv'
    rb.analyzer = MockAnalyzer
    rb.display({})
    assert True # no exception


  def test_email(self, mocker):
    # assume user is not accepting to share by email, since will be prompted
    mockee = 'click.confirm'
    mocker.patch(mockee, side_effect=lambda x: False)
    #mockee = 'click.prompt'
    #mocker.patch(mockee, side_effect=lambda x: 'whatever')

    import pytest
    rb = ReporterOptimize()
    context_all = rb.email({'emailTo': []})
    assert True # assert no exceptions. The above .email will just return silently without doing anything since it is not implemented



from isitfit.cost.account_cost_analyze import ServiceReporterBinned

class TestServiceReporterBinned:
  # pytest isitfit/tests/cost/redshift/test_reporter.py::TestReporterAnalyze::test_email
  def test_email(self, mocker):
    mockee = 'isitfit.emailMan.EmailMan'
    mocker.patch(mockee, autospec=True)

    mockee = 'isitfit.cost.base_reporter.ReporterBase.email'
    #mockee = 'isitfit.cost.account_cost_analyze.ReporterBase'
    # mocker.patch(mockee, autospec=True)
    mocker.patch(mockee, side_effect=lambda x: None)


    class MockClickCtx:
      obj = {
        'aws_profile': 'bla',
        'filter_region': None
      }

    rb = ServiceReporterBinned()
    rb.emailTo = []
    import pandas as pd
    context_all = {'dfbin_p': pd.DataFrame(), 'click_ctx': MockClickCtx()}
    rb.email(context_all)
    assert True # no exception


  def test_display(self):
    import datetime as dt
    dt_now = dt.datetime.utcnow()

    class MockAnalyzer:
      class MockIter:
        rc_noData = []
      class MockCw:
        pass

      rp_iter = MockIter
      cwman = MockCw

    rb = ServiceReporterBinned()
    context_all = {
      'dfbin_p': []
    }
    rb.display(context_all)
    assert True # no exception




import click


def pingOnError(ctx, error):
  # if not dict yet, i.e. before the cli.core.cli_core group
  # Important to keep track of the didPing variable
  if ctx.obj is None: return

  # check if the error's ping was done
  didPing = 'unhandled_error_pinged' in ctx.obj.keys()
  if didPing: return

  # send to sentry.io via isitfit.io (check usage of sentry_proxy in cli.core)
  from sentry_sdk import capture_exception
  capture_exception(error)

  # proceed to ping matomo about the error (to be deprecated in full in favor of sentry)
  from isitfit.utils import ping_matomo
  exception_type = type(error).__name__ # https://techeplanet.com/python-catch-all-exceptions/
  exception_str = ""
  try:
    exception_str = str(error)
  except:
    pass

  ping_matomo("/error/unhandled/%s?message=%s"%(exception_type, exception_str))

  # save a flag saying that the error sent a ping
  # Note that it is not necessary to do more than that, such as storing a list of pinged errors,
  # because there will be exactly one error raise at most before the program fails
  ctx.obj['unhandled_error_pinged'] = True


class IsitfitGroup(click.Group):
  """
  Wraps the click.Group.invoke function to ping matomo about the error before bubbling all exceptions
  """
  def invoke(self, ctx):
    try:
      ret = super().invoke(ctx)
      return ret
    except Exception as error:
      pingOnError(ctx, error)
      raise


from isitfit.utils import display_footer

class IsitfitCommand(click.Command):
  """
  Wraps the click.Command.invoke function to ping matomo about the error before bubbling all exceptions

  Also Call display_footer at the end of each invokation
    https://github.com/pallets/click/blob/8df9a6b2847b23de5c65dcb16f715a7691c60743/click/core.py#L945
  """
  def invoke(self, ctx):
    try:
      ret = super().invoke(ctx)
      display_footer()
      return ret
    except Exception as error:
      pingOnError(ctx, error)
      raise


def isitfit_group(name=None, **attrs):
  """
  Overrides click.decorators.group to use the class IsitfitGroup
  """
  attrs.setdefault('cls', IsitfitGroup)
  return click.command(name, **attrs)


# Inherit from click's usageError since click can handle it automatically
# https://click.palletsprojects.com/en/7.x/exceptions/
class IsitfitCliError(click.UsageError):
  """
  Inherited from click.exceptions.UsageError
  because it adds the context as a constructor argument,
  which I need for checking "is_outdated"
  https://github.com/pallets/click/blob/8df9a6b2847b23de5c65dcb16f715a7691c60743/click/exceptions.py#L51
  """

  # exit code
  exit_code = 10

  # constructor parameters from
  # https://github.com/pallets/click/blob/8df9a6b2847b23de5c65dcb16f715a7691c60743/click/exceptions.py#L11
  def show(self, file=None):
    # ping matomo about error
    from isitfit.utils import ping_matomo
    ping_matomo("/error?message=%s"%self.message)

    # continue
    from click._compat import get_text_stderr
    if file is None:
        file = get_text_stderr()

    # echo wrap
    color = 'red'
    def wrapecho(message):
      # from click.utils import echo
      # echo('Error: %s' % self.format_message(), file=file, color=color)

      click.secho(message, fg=color)

    # main error
    wrapecho('Error: %s' % self.format_message())

    # if error from terminal during execution (not on program boot)
    if self.ctx is not None:
      if self.ctx.obj is not None:
        # if isitfit installation is outdated, append a message to upgrade
        if self.ctx.obj.get('is_outdated', None):
          hint_1 = "Upgrade your isitfit installation with `pip3 install --upgrade isitfit` and try again."
          wrapecho(hint_1)

        # test that boto3 minimum command can run
        # This would fail for example for: `AWS_ACCESS_KEY_ID=wrong AWS_SECRET_ACCESS_KEY=alsowrong aws iam get-user`
        import boto3
        iam_client = boto3.client('iam')
        try:
            # response = iam_client.get_user()
            iam_client.get_user()
        except Exception as e:
            msg_e = str(e)
            hint_3 = f"""Hint: The command `aws iam get-user` has also failed with the following error:
      {msg_e}
      This might indicate a problem with your aws user's permissions and could be related to the current error in isitfit."""
            # Update 2020-01-09 Instead of raising an exception, just display a warning
            #from isitfit.cli.click_descendents import IsitfitCliError
            #raise IsitfitCliError(hint_3) from e

            # ping matomo about warning
            from isitfit.utils import ping_matomo
            ping_matomo("/warning/aws-iam-get-user?message=%s"%hint_3)

            # display on screen
            wrapecho(hint_3)


    # add link to github issues
    hint_2 = "If the problem persists, please report it at https://github.com/autofitcloud/isitfit/issues/new"
    wrapecho(hint_2)



def isitfit_option_base(name=None, **attrs):
  """
  Overrides click.option due to trouble with options having prompt=True (or prompt='bla') and the --help call
  eg `isitfit cost analyze --help` was prompting for `--profile`
     `isitfit version` was also prompting for `--profile` when it was at the `cli_core` level

  This resolves click github issue: https://github.com/pallets/click/issues/295

  Original notes (no longer valid since problems solved with isitfit_option):
    --profile docs at
      https://click.palletsprojects.com/en/7.x/options/?highlight=prompt#values-from-environment-variables
    The first commented version with prompt=... is not possible to implement because in this case, `isitfit cost analyze --help` will also prompt for the profile
      (similar to the --ndays reason it was duplicated to the click.commands instead of keeping it at the click.group leve)
    Also, originally, this was in the cli.core.cli_core group,
      and eventhough Click has a "eager" option and even a "version_option", but I'm not using those ATM
      and the problem there was with `isitfit version` prompting for this.
    This time, with profile, unlike ndays, instead of duplicating it for each command,
      I'm implementing it manually if not provided
  """
  if not attrs.get('prompt'):
    # nothing special required
    return click.option(name, **attrs)

  # shortcuts
  callback_ori = attrs.get('callback')
  prompt_ori = attrs.get('prompt')
  default_ori = attrs.get('default')
  type_ori = attrs.get('type')

  # wrap callback_ori with manual prompt IF NO --help
  def callback_wrap(ctx, param, value):
    # check for --help
    import sys
    if '--help' in sys.argv:
      # no need to do anything
      return value

    if value is None:
      # manually prompt only if not already supplied
      value = click.prompt(prompt_ori, default=default_ori, type=type_ori)
    else:
      value = type_ori(value)

    # call the original callback
    if callback_ori is None: return value
    return callback_ori(ctx, param, value)

  # over-ride the originals since moved into callback_wrap
  attrs['callback'] = callback_wrap
  attrs['prompt'] = False
  attrs['default'] = None
  attrs['type'] = None

  # return the "corrected" option
  return click.option(name, **attrs)



def isitfit_option_profile(name=None, **attrs):
  """
  Special option dedicated to --profile
  Was worth it because it was in cli.tags and cli.cost
  """
  from isitfit.utils import AwsProfileMan
  profile_man = AwsProfileMan()

  ret_opt = isitfit_option_base('--profile', type=str, default=profile_man.default(), callback=profile_man.validate_profile, prompt=profile_man.prompt(), help='Use a specific profile from your credential file.', envvar='AWS_PROFILE')
  return ret_opt




from .metrics_datadog import HostNotFoundInDdg, DataNotFoundForHostInDdg
from .metrics_cloudwatch import NoCloudwatchException
from isitfit.utils import logger

class MetricsAuto:
  """
  Auto-fallback metrics fetching:
  If datadog credentials set in environment variable,
  try to get the metrics for a resource from datadog,
  otherwise fall back to cloudwatch
  """
  def __init__(self, datadog, cloudwatch):
    # class of instance metrics_datadog.DatadogCached
    self.datadog = datadog

    # class of instance metrics_cloudwatch.CloudwatchEc2 or CloudwatchRedshift
    self.cloudwatch = cloudwatch

    # dict with keys being host ID and value being
    # another dict with keys: ID, datadog, cloudwatch
    # and values being: host ID, status of fetching data from datadog, status of fetching data from cloudwatch
    self.status = {}


  def _try_datadog(self, aws_id):
    if not self.datadog.is_configured():
      return None, "not configured"

    try:
      df_ddg = self.datadog.get_metrics_all(aws_id)
      return df_ddg, "ok"
    except HostNotFoundInDdg as e:
      logger.debug("Datadog: host not found for aws ID %s: %s"%(aws_id, str(e)))
      return None, "host not found"
    except DataNotFoundForHostInDdg as e:
      logger.debug("Datadog: data not found for aws ID %s: %s"%(aws_id, str(e)))
      return None, "no data"


  def _try_cloudwatch(self, host_id, host_region, host_created):
    try:
      df_cw  = self.cloudwatch.handle_main({'Region': host_region}, host_id, host_created)
      return df_cw, "ok"
    except NoCloudwatchException:
      logger.debug("Cloudwatch: data not found for %s"%host_id)
      return None, "no data"


  def set_ndays(self, ndays):
    self.cloudwatch.set_ndays(ndays)
    self.datadog.set_ndays(ndays)


  def handle_host(self, host_id, host_region, host_created):
    logger.debug("host id, region, created: %s, %s, %s"%(host_id, host_region, host_created))

    self.status[host_id] = {
      'ID': host_id,
      'datadog': 'Did not try',
      'cloudwatch': 'Did not try'
    }
    df, status = self._try_datadog(host_id)
    self.status[host_id]['datadog'] = status
    if status!='ok':
      # "df is None" and status!=ok are equivalent
      df, status  = self._try_cloudwatch(host_id, host_region, host_created)
      self.status[host_id]['cloudwatch'] = status

    return df


  def display_status(self):
    # choose main function to display
    #from isitfit.utils import logger
    #display_msg = logger.info
    #import click
    #display_msg = click.echo

    # warning + color yellow
    from isitfit.utils import logger
    from termcolor import colored
    #display_msg = lambda x: colored(logger.warning(x), 'yellow')
    import click
    display_msg = lambda x: click.secho(x, fg='yellow')

    # inferred service
    from isitfit.cost.metrics_cloudwatch import CloudwatchEc2, CloudwatchRedshift
    service_map = {
      CloudwatchEc2: 'ec2',
      CloudwatchRedshift: 'redshift'
    }
    service_inferred = service_map.get(type(self.cloudwatch), 'unknown')

    # to dataframe
    import pandas as pd
    df = pd.DataFrame(self.status.values())

    if df.shape[0]==0:
      display_msg("")
      display_msg("Status of metric sources (for service %s)"%service_inferred)
      display_msg("<Empty table>")
      display_msg("")
      return

    df = df.sort_values(['ID'])
    gp = df.groupby(['datadog', 'cloudwatch']).count() # .reset_index().rename(columns={'ID': 'n'})
    from isitfit.dotMan import DotMan
    tempdir = DotMan().tempdir()
    import tempfile
    fh = tempfile.NamedTemporaryFile(prefix="isitfit-sourceStatus-", suffix='.csv', dir=tempdir, delete=False)
    df.to_csv(fh.name, index=False)

    # print with tabulate
    from tabulate import tabulate
    display_msg("")
    display_msg("Status of metric sources (for service %s)"%service_inferred)
    display_msg(tabulate(gp.reset_index(), tablefmt='psql', showindex=False, headers='keys'))
    display_msg("Details at %s"%fh.name)
    display_msg("")



class MetricsListener(MetricsAuto):
  """
  Listener for event bus defined in mainManager.py
  """
  def per_host(self, context_host):
    host_id = context_host['ec2_id']
    host_region, host_created = None, None

    ec2_obj = context_host['ec2_obj']
    if ec2_obj is not None:
      host_region = ec2_obj.region_name
      host_created = ec2_obj.launch_time

    df = self.handle_host(host_id, host_region, host_created)

    if df is None:
      # if still no data, just break the chain of listeners
      return None

    # otherwise if data is available
    context_host['df_metrics'] = df
    return context_host

  def display_status(self, context_all):
    super().display_status()
    return context_all



from isitfit.utils import logger

from isitfit.cost.base_reporter import ReporterBase

class ServiceAggregator:
  def __init__(self):
    self.table_d = {'ec2': {}, 'redshift': {}}
    self.table_c = None
    self.table_a = None

  def per_service_save(self, context_service):
    service_name = context_service['ec2_id']
    context_all = context_service['context_all']
    if context_all is None: return context_service

    if service_name=='ec2':
      self.table_d['ec2']['df_sort'] = context_all['df_sort']
      self.table_d['ec2']['sum_val'] = context_all['sum_val']
      # after migration from display to display2, no need for these
      # self.table_d['ec2']['csv_fn_final'] = context_all['csv_fn_final']
      # self.table_d['ec2']['analyzer'] = context_all['analyzer']

    elif service_name=='redshift':
      self.table_d['redshift']['analyzer'] = context_all['analyzer']
      # after migration from display to display2, no need for these
      # self.table_d['redshift']['analyze_df'] = context_all['analyzer'].analyze_df
      # self.table_d['redshift']['csv_fn_final'] = context_all['csv_fn_final']

    else:
      raise Exception("Invalid service runner description: %s"%service_name)

    return context_service


  def _concat_ec2(self):
    if 'df_sort' not in self.table_d['ec2']:
      return None

    if self.table_d['ec2']['df_sort'] is None:
      return None

    # get 2 dataframes
    t_ec2 = self.table_d['ec2']['df_sort'].copy()

    # add service column
    t_ec2['service'] = 'EC2'

    # rename columns to match
    t_ec2.rename(columns={'instance_id':'resource_id', 'instance_type':'resource_size1', 'recommended_type':'recommended_size1'}, inplace=True)

    # append columns that exist in redshift, but not in ec2
    t_ec2['resource_size2'] = None

    return t_ec2


  def _concat_redshift(self):
    if 'analyzer' not in self.table_d['redshift'].keys():
      return

    if self.table_d['redshift']['analyzer'].analyze_df.shape[0]==0:
      return

    # get 2 dataframes
    t_rsh = self.table_d['redshift']['analyzer'].analyze_df.copy()

    # add service column
    t_rsh['service'] = 'Redshift'

    # rename columns to match
    t_rsh.rename(columns={'Region':'region', 'ClusterIdentifier':'resource_id', 'NodeType':'resource_size1', 'NumberOfNodes':'resource_size2', 'classification':'classification_1'}, inplace=True)
    del t_rsh['CpuMaxMax']
    del t_rsh['CpuMinMin']

    return t_rsh


  def concat(self, context_all):
    # get 2 dataframes
    t_ec2 = self._concat_ec2()
    t_rsh = self._concat_redshift()

    # concatenate
    t_all = [t_ec2, t_rsh]
    t_all = [x for x in t_all if x is not None]

    if len(t_all)==0:
      context_all['table_c'] = None
      return context_all

    import pandas as pd
    self.table_c = pd.concat(t_all, axis=0, sort=False)

    # order columns
    # self.table_c.set_index([], inplace=True) # do not set index since display_df ignores the index ATM
    cols_theo = ['service', 'region', 'resource_id', 'resource_size1', 'resource_size2', 'classification_1', 'classification_2', 'cost_3m', 'recommended_size1', 'savings', 'tags']

    # the "if x in" part is needed because if there are no redshift recommendations, there would be no column resource_size2
    # Update 2019-12-26 no longer need this since adding resource_size2=None to ec2 results. Keeping it anyway for now
    cols_all = [x for x in cols_theo if x in self.table_c.columns]

    # actual ordering of columns
    self.table_c = self.table_c[cols_all]

    # group by for summary
    # TODO
    #self.table_a = self.table_c.groupby(['service', 'region'])

    # add a new dt_detected for all recommendations. Those which had a dt_detected generated earlier will have this overwritten later
    import datetime as dt
    import pytz
    self.table_c['dt_detected'] = dt.datetime.utcnow().replace(tzinfo=pytz.utc)

    # save to context
    context_all['table_c'] = self.table_c

    return context_all



class SqliteMan:
  def __init__(self, click_ctx):
    # get profile name (for per-profile sqlite databases)
    # Note, no need to do a per-ndays database, since a recommendation of "change size from S1 to S2" is the same whether it is based on 7 days or 90 days
    # profile_name = context_all['click_ctx'].obj['aws_profile']
    profile_name = click_ctx.obj['aws_profile']

    # settings for local sqlite database
    import sqlite3
    from isitfit.dotMan import DotMan
    import os
    self.db_path = os.path.join(DotMan().get_dotisitfit(), "db_cost_optimize_01-%s.sqlite"%profile_name)
    #self.db_exists = os.path.exists(self.db_path) # before the .connect below automatically creates the file
    self.db_conn = sqlite3.connect(self.db_path)


  def read_sqlite(self, context_pre):
    # No need since the read_sql_query below will fail if the database didnt already exist
    #if not self.db_exists:
    #  return context_pre

    import pandas as pd
    try:
      table_previous = pd.read_sql_query('select * from recommendations_previous', self.db_conn)
    #except sqlite3.OperationalError:
    except pd.io.sql.DatabaseError:
      return context_pre

    # un-encode the tags_json field
    import json
    table_previous['tags'] = table_previous.tags_json.apply(lambda x: json.loads(x))
    del table_previous['tags_json']

    # display
    reporter = ServiceReporter()
    reporter.display2({'table_c': table_previous})

    # prompt user if should re-calculate, will raise click.Abort if user answers No
    # https://click.palletsprojects.com/en/7.x/api/?highlight=click%20confirm#click.confirm
    from termcolor import colored
    import click
    click.confirm(colored("Do you want to recalculate?", "red"), default=False, abort=True)

    # do not return into context since the program flow is:
    # - either abort now
    # - or re-calculate table_c in the aggregator
    #context_pre['table_c'] = table_previous
    return context_pre


  def update_dtCreated(self, context_all):
    """
    Save results to sqlite database, so that recommendations can get a UID attached and get tracked between re-runs
    """
    table_current = context_all['table_c']
    if table_current is None:     return context_all
    #if table_current.shape[0]==0: return context_all

    # convert column "tags" to "tags_json" since it is a list, and sqlite doesn't support lists
    import json
    table_current['tags_json'] = table_current.tags.apply(lambda x: json.dumps(x))

    # columns that need to go to sqlite
    cols_sqlite = [
      'service', 'region', 'resource_id',
      'resource_size1', 'resource_size2', 'classification_1', 'classification_2',
      'cost_3m', # not needed in left join below, but needed for read_sqlite
      'recommended_size1',
      'savings', # not needed in left join below, but needed for read_sqlite
      # 'tags', # This is a list, which sqlite doesn't support, so use tags_json instead (json-encoded)
      'tags_json', # not needed in left join below, but needed for read_sqlite
      'dt_detected',
    ]
    # cols_sqlite = [x for x in cols_sqlite if x in table_current.columns] # Update 2019-12-26 no longer need this since adding resource_size2=None to ec2 results

    # make sure that the table recommendations_previous exists
    # (use trick to append a 0-row table)
    df_empty = table_current[cols_sqlite].iloc[0:0]
    assert df_empty.shape[0] == 0
    df_empty.to_sql('recommendations_previous', self.db_conn, if_exists='append', index=False)

    # insert current table of recommendations into db
    table_current[cols_sqlite].to_sql('recommendations_current', self.db_conn, if_exists='replace', index=False)

    # join on previously saved recommendations
    # Note: the table recommendations_previous = recommendations_current with the dt_detected being the 1st date on which it was generated
    # Also note that the result is just the current recommendations
    import pandas as pd
    sql = """
    select
      -- service + region + resource ID: unique over resource life, other fields changeable
      rc.service,
      rc.region,
      rc.resource_id,

      -- date this recommendation was created
      COALESCE(rp.dt_detected, rc.dt_detected) as dt_detected

    from recommendations_current as rc
    left join recommendations_previous as rp
       on rc.service           = rp.service
      and rc.region            = rp.region
      and rc.resource_id       = rp.resource_id
      and rc.resource_size1    = rp.resource_size1
      and coalesce(rc.resource_size2, 'na')    = coalesce(rp.resource_size2, 'na') -- field that is missing for ec2
      and rc.classification_1  = rp.classification_1
      and rc.classification_2  = rp.classification_2
      and rc.recommended_size1 = rp.recommended_size1
    """
    table_dated = pd.read_sql_query(sql=sql, con=self.db_conn)

    # merge back table_dated into table_current (to get the corrected dt_detected field)
    cols_exDate = [x for x in table_current.columns if x!='dt_detected']
    table_current = table_current[cols_exDate].merge(table_dated, how='left', on=['service', 'region', 'resource_id'])

    # save table with corrected dt_detected back into db
    # Note that I'm not sure how this plays out with the --n=1 option
    # Method 1: simple, but loses all recommendations instead of preserving them all
    #           i.e. if a recommendation set is missing a recommendation R1, then it is deleted
    table_current[cols_sqlite].to_sql('recommendations_previous', self.db_conn, if_exists='replace', index=False)

    # drop the tags_json field since not needed except for sqlite
    del table_current['tags_json']

    # update in context
    context_all['table_c'] = table_current

    # done
    return context_all


class ServiceReporter(ReporterBase):
  # commenting this out in favor of display2
  # def display(self, context_all):
  #   # ATM just using the individual service reports
  #   from isitfit.cost.ec2_optimize import ReporterOptimizeEc2
  #   roe = ReporterOptimizeEc2()
  #   if 'df_sort' not in self.table_d['ec2']:
  #     import click
  #     click.echo("No optimizations from EC2")
  #   else:
  #     roe.df_sort = self.table_d['ec2']['df_sort']
  #     roe.sum_val = self.table_d['ec2']['sum_val']
  #     roe.csv_fn_final = self.table_d['ec2']['csv_fn_final']
  #     roe.analyzer = self.table_d['ec2']['analyzer']
  #     roe.display(context_all)
  #
  #   from isitfit.cost.redshift_optimize import ReporterOptimize as ReporterOptimizeRedshift
  #   ror = ReporterOptimizeRedshift()
  #   if 'analyzer' not in self.table_d['redshift'].keys():
  #     import click
  #     click.echo("No optimizations from redshift")
  #   elif self.table_d['redshift']['analyzer'].analyze_df.shape[0]==0:
  #     import click
  #     click.echo("No optimizations from redshift")
  #   else:
  #     ror.analyzer = self.table_d['redshift']['analyzer']
  #     ror.csv_fn_final = self.table_d['redshift']['csv_fn_final']
  #     ror.display(context_all)
  #
  #   return context_all


  def display2(self, context_all):
    """
    Re-write of self.display to merge the 2 dataframes of EC2 and Redshift into one table
    """
    self.table_c = context_all['table_c']

    # ATM just using the individual service reports
    import click

    # if 'df_sort' not in self.table_d['ec2']:
    if self.table_c is None:
      click.secho("No optimizations from EC2", fg='red')
    elif 'EC2' not in self.table_c['service'].to_list():
      click.secho("No optimizations from EC2", fg='red')
    else:
      # copy from ec2_optimize.Reporter.display
      # sum_val = self.table_d['ec2']['sum_val']
      sum_val = self.table_c.savings.sum()
      if sum_val==0:
        # Update 2019-12-12 bring back the echo below (after having been commented out for a few days)
        click.secho("No optimizations from EC2", fg='red')
        #pass
      elif sum_val is None:
        # Update 2019-12-12 bring back the echo below (after having been commented out for a few days)
        click.secho("No optimizations from EC2", fg='red')
        #pass
      else:
        sum_comment = "extra cost" if sum_val>0 else "savings"
        sum_color = "red" if sum_val>0 else "green"
        click.secho("EC2 %s over the next 3 months: $ %0.0f"%(sum_comment, sum_val), fg=sum_color)

    # spacer
    click.echo("")

    #if 'analyzer' not in self.table_d['redshift'].keys():
    #  click.secho("No optimizations from redshift", fg='red')
    #elif self.table_d['redshift']['analyzer'].analyze_df.shape[0]==0:
    #  click.secho("No optimizations from redshift", fg='red')

    if self.table_c is None:
      click.secho("No optimizations from redshift", fg='red')
    elif 'Redshift' not in self.table_c['service'].to_list():
      click.secho("No optimizations from redshift", fg='red')

    if self.table_c is None:
      return context_all

    if self.table_c.shape[0]==0:
      return context_all

    # save concatenated table to CSV
    # copied from isitfit.cost.optimizationListener.storecsv...
    import tempfile
    from isitfit.dotMan import DotMan
    with tempfile.NamedTemporaryFile(prefix='isitfit-costOptimize-', suffix='.csv', delete=False, dir=DotMan().tempdir()) as csv_fh_final:
      click.secho("Saving final results to %s"%csv_fh_final.name, fg="cyan")
      self.table_c.to_csv(csv_fh_final.name, index=False)
      click.secho("Save complete", fg="cyan")

    # display concatenated table
    from isitfit.utils import display_df
    display_df(
      "Optimization summary",
      self.table_c,
      # self.table_a, # TODO
      csv_fh_final.name,
      self.table_c.shape,
      logger
    )

    # done
    return context_all



def pipeline_factory(mm_eco, mm_rco, ctx):
    from isitfit.cost.mainManager import RunnerAccount
    mm_all = RunnerAccount("AWS cost optimize (EC2, Redshift) in all regions", ctx)

    # add listener that checks the local sqlite database for a previous calculation
    # and display if available, then prompt user if desires to re-calculate
    # Note that if no desire to re-calculate, this raises an exception that bubbles up into get_ifi and aborts the pipeline early
    sqlite_man = SqliteMan(ctx)

    # Update 2019-12-27 will not display the database table ATM, in favor of using it in the interactive command
    # mm_all.add_listener('pre', sqlite_man.read_sqlite)

    # set up a pipeline that fetches fresh data
    from .account_cost_analyze import ServiceIterator, ServiceCalculatorGet
    iterator = ServiceIterator(mm_eco, mm_rco)
    mm_all.set_iterator(iterator)

    # ping matomo at start of calculation
    from isitfit.utils import ping_matomo
    inject_timer_start = lambda context_pre: context_pre if ping_matomo("/cost/optimize/account/start") else context_pre
    mm_all.add_listener('pre', inject_timer_start)

    calculator_get = ServiceCalculatorGet()
    mm_all.add_listener('ec2', calculator_get.per_service)

    aggregator = ServiceAggregator()
    mm_all.add_listener('ec2', aggregator.per_service_save)
    mm_all.add_listener('all', aggregator.concat)

    mm_all.add_listener('all', sqlite_man.update_dtCreated)

    # ping matomo at end of calculation
    inject_timer_end = lambda context_all: context_all if ping_matomo("/cost/optimize/account/end") else context_all
    mm_all.add_listener('all', inject_timer_end)

    # whether reading from sqlite or fresh data, display
    reporter = ServiceReporter()
    #mm_all.add_listener('all', reporter.display)
    mm_all.add_listener('all', reporter.display2)

    # done
    return mm_all



from isitfit.utils import logger






def get_dtNowD():
  import datetime as dt
  import pytz
  dt_now_d = dt.datetime.utcnow().replace(tzinfo=pytz.utc)
  return dt_now_d


import math
dt_now_d = get_dtNowD()

from isitfit.utils import mergeSeriesOnTimestampRange


# convert above dict to pandas dataframe
from isitfit.cost.catalog_redshift import redshiftPricing_df



from isitfit.cost.redshift_common import CalculatorBaseRedshift
class CalculatorAnalyzeRedshift(CalculatorBaseRedshift):

  def per_ec2(self, context_ec2):
      # parent
      context_ec2 = super().per_ec2(context_ec2)

      # unpack
      rc_describe_entry, rc_id, rc_created = context_ec2['ec2_dict'], context_ec2['ec2_id'], context_ec2['ec2_launchtime']

      # pandas series of number of nodes available, and node type, over time, past 90 days
      df_type_ts1 = context_ec2['df_type_ts1']
      df_type_ts1 = df_type_ts1.rename(columns={'ResourceSize1': 'NodeType', 'ResourceSize2': 'NumberOfNodes'})

      # get all performance dataframes, on the cluster-aggregated level
      df_single = context_ec2['df_single']

      rc_type = rc_describe_entry['NodeType']

      # append a nhours field per day
      # and correct for number of hours on first and last day
      # Note that intermediate days are just 24 hours since Redshift cannot be stopped
      # Update 2019-11-20
      # FIXME need to look at how to use SampleCount for this to get a more accurate number,
      # especially for days on which an instance is run for the first 10 hours then deleted
      # To note that initially I had dismissed this SampleCount idea because a redshift cluster cannot be
      # stopped and restarted like ec2, but a major difference is the ability to re-assign a past ID to the cluster
      # unlike ec2 where the instance ID gets generated randomly for each terminated and re-created instance
      # eg
      # (Pdb) df_single.head()
      #     Timestamp  SampleCount   Average   Minimum    Maximum     Unit  nhours
      # 2  2019-11-19        918.0  2.324190  0.833333  53.500000  Percent       24
      # 0  2019-11-20          4.0  1.208333  1.000000   1.416667  Percent        4
      #
      # Update 2019-12-17 It turns out SampleCount for redshift is 1 per 10 minutes (wherease EC2 is 1 per 5 minutes)
      # Commenting out the nhours calculation below and no need to add new calculation because Cloudwatch metrics class will already add the nhours field
      #ymd_creation = rc_describe_entry['ClusterCreateTime'].strftime("%Y-%m-%d")
      #ymd_today    = dt_now_d.strftime("%Y-%m-%d")
      #
      #hc_ref = dt_now_d if ymd_creation == ymd_today else rc_describe_entry['ClusterCreateTime'].replace(hour=23, minute=59)
      #hours_creation = hc_ref - rc_describe_entry['ClusterCreateTime']
      #hours_creation = math.ceil(hours_creation.seconds/60/60)
      #hours_today = dt_now_d - dt_now_d.replace(hour=0, minute=0)
      #hours_today = math.ceil(hours_today.seconds/60/60)
      #
      #def calc_nhours(ts):
      #  ts_str = ts.strftime("%Y-%m-%d")
      #  if ts_str == ymd_creation: return hours_creation
      #  if ts_str == ymd_today:    return hours_today
      #  return 24
      #
      #df_single['nhours'] = df_single.Timestamp.apply(calc_nhours)

      # merge df_single (metrics) with df_type_ts1 (cloudtrail history)
      # (adds column NodeType, NumberOfNodes)
      df_single = mergeSeriesOnTimestampRange(df_single, df_type_ts1, ['NodeType', 'NumberOfNodes'])

      # merge with the price catalog (adds column Cost)
      df_single = df_single.merge(redshiftPricing_df, left_on='NodeType', right_on='NodeType', how='left')

      # calculate columns capacity_usd and used_usd
      df_single['used_usd'    ] = df_single.cpu_used_avg / 100 * df_single.Cost * df_single.nhours * df_single.NumberOfNodes
      df_single['capacity_usd'] = 1                       * df_single.Cost * df_single.nhours * df_single.NumberOfNodes

      # save back to context for further binning
      context_ec2['df_single'] = df_single

      # summarize into 1 row
      self.analyze_list.append({
        'ClusterIdentifier': rc_describe_entry['ClusterIdentifier'],

        # most recent node type and number of nodes
        'NodeType': rc_type,
        'NumberOfNodes': rc_describe_entry['NumberOfNodes'],
        'Region': rc_describe_entry['Region'],

        # cost used/billed, including cloudtrail history
        'CostUsed':   df_single['used_usd'    ].sum(),
        'CostBilled': df_single['capacity_usd'].sum()
      })

      # done
      return context_ec2



  def calculate(self, context_all):
    # defaults
    self.cost_used = 0
    self.cost_billed = 0
    self.cwau_percent = 0
    self.regions_n = 0

    # calculate cost-weighted utilization
    analyze_df = self.analyze_df

    # no data, eg when ndays=1 and a cluster is launched a few minutes ago,
    # cloudwatch has a bug whereby it doesn't return today's data anymore.
    # Need to set ndays>=64 to get today's data. Check related bugfix in ec2_analyze
    if analyze_df.shape[0]==0:
      return context_all

    self.cost_used   = analyze_df.CostUsed.fillna(value=0).sum()
    self.cost_billed = analyze_df.CostBilled.fillna(value=0).sum()
    self.regions_n = len(analyze_df.Region.unique())

    if self.cost_billed == 0:
      self.cwau_percent = 0
      return context_all

    self.cwau_percent = int(self.cost_used / self.cost_billed * 100)
    return context_all





from isitfit.cost.base_reporter import ReporterBase
class ReporterAnalyze(ReporterBase):
  def postprocess(self, context_all):
    # unpack
    self.analyzer = context_all['analyzer']
    mm = context_all['mainManager']
    n_rc_total = context_all['n_ec2_total']
    n_rc_analysed = context_all['n_rc_analysed']

    # copied from isitfit.cost.ec2.calculator_analyze.after_all
    cwau_val = self.analyzer.cwau_percent
    cwau_color = 'yellow'
    if cwau_val >= 70: cwau_color = 'green'
    elif cwau_val <= 30: cwau_color = 'red'

    dt_start = mm.StartTime.strftime("%Y-%m-%d")
    dt_end   = mm.EndTime.strftime("%Y-%m-%d")

    self.table = [
      { 'color': '',
        'label': "Start date",
        'value': "%s"%dt_start
      },
      { 'color': '',
        'label': "End date",
        'value': "%s"%dt_end
      },
      { 'color': '',
        'label': "Regions",
        'value': "%i"%self.analyzer.regions_n
      },
      { 'color': '',
        'label': "Redshift clusters (total)",
        'value': "%i"%n_rc_total
      },
      { 'color': '',
        'label': "Redshift clusters (analyzed)",
        'value': "%i"%n_rc_analysed
      },
      { 'color': '',
        'label': "Billed cost",
        'value': "%0.0f $"%self.analyzer.cost_billed
      },
      { 'color': '',
        'label': "Used cost",
        'value': "%0.0f $"%self.analyzer.cost_used
      },
      { 'color': cwau_color,
        'label': "CWAU (Used/Billed)",
        'value': "%0.0f %%"%cwau_val
      },
    ]

    # save in context for aggregator
    context_all['table'] = self.table

    # done
    return context_all

# DEPRECATED in favor of the reporter in account_cost_analyze.py
#
#  def display(self, context_all):
#    # copied from isitfit.cost.ec2.calculator_analyze.display_all
#
#    # https://pypi.org/project/termcolor/
#    from termcolor import colored
#
#    def get_row(row):
#        def get_cell(i):
#          retc = row[i] if not row['color'] else colored(row[i], row['color'])
#          return retc
#
#        retr = [get_cell('label'), get_cell('value')]
#        return retr
#
#    dis_tab = [get_row(row) for row in self.table]
#
#    from tabulate import tabulate
#
#    # logger.info("Summary:")
#    logger.info("Cost-Weighted Average Utilization (CWAU) of the AWS Redshift account:")
#    logger.info("")
#    logger.info(tabulate(dis_tab, headers=['Field', 'Value']))
#    logger.info("")
#    logger.info("For reference:")
#    logger.info(colored("* CWAU >= 70% is well optimized", 'green'))
#    logger.info(colored("* CWAU <= 30% is underused", 'red'))
#    logger.info("")
#    logger.info("For the EC2 analysis, scroll up to the previous table.")
#    return context_all
#
#
#  def email(self, context_all):
#      context_2 = {}
#      context_2['emailTo'] = context_all['emailTo']
#      context_2['click_ctx'] = context_all['click_ctx']
#      context_2['dataType'] = 'cost analyze' # redshift, not ec2
#      context_2['dataVal'] = {'table': self.table}
#      super().email(context_2)
#
#      return context_all




def pipeline_factory(share_email, filter_region, ctx, filter_tags):
  # This is a factory method, so it doesn't make sense to display "Analyzing bla" if actually "foo" is analyzed first
  #logger.info("Analyzing redshift clusters")

  from .redshift_common import redshift_cost_core
  ra = CalculatorAnalyzeRedshift()
  rr = ReporterAnalyze()
  mm = redshift_cost_core(ra, rr, share_email, filter_region, ctx, filter_tags)
  return mm




import click
from isitfit.cli.click_descendents import IsitfitCommand, isitfit_group

@isitfit_group(help="Manage migrations for local files (useful for debugging)", invoke_without_command=False, hidden=True)
@click.pass_context
def migrations(ctx):
  # FIXME click bug: `isitfit command subcommand --help` is calling the code in here. Workaround is to check --help and skip the whole section
  import sys
  if '--help' in sys.argv: return

  # usage stats
  from isitfit.utils import ping_matomo
  ping_matomo("/migrations")

  from isitfit.migrations.migman import MigMan
  migman = MigMan()
  migman.connect()
  migman.read()

  ctx.obj['migman'] = migman


@migrations.command(help="Show all migrations", cls=IsitfitCommand)
@click.pass_context
def show(ctx):
  # usage stats
  from isitfit.utils import ping_matomo
  ping_matomo("/migrations/show")

  migman = ctx.obj['migman']

  if migman.df_mig.shape[0]==0:
    click.echo("No pending migrations")
  else:
    click.echo("Pending migrations")
    click.echo(migman.df_mig[['migname', 'description']])
    click.echo("")
    click.secho("Use `isitfit migrations migrate` to execute them", fg="yellow")


@migrations.command(help="Execute pending migrations", cls=IsitfitCommand)
@click.option('--not-dry-run', is_flag=True, help='Simulate the migration without executing it')
@click.pass_context
def migrate(ctx, not_dry_run):
  # usage stats
  from isitfit.utils import ping_matomo, b2l
  ping_matomo("/migrations/migrate?not_dry_run=%s"%b2l(not_dry_run))

  migman = ctx.obj['migman']
  migman.not_dry_run = not_dry_run
  migman.migrate_all()

  if not not_dry_run:
    click.echo("")
    click.secho("This was a simulated execution", fg="yellow")
    click.secho("Repeat using `isitfit migrations migrate --not-dry-run` for actual execution", fg='yellow')



# imports
import datetime as dt
from isitfit.utils import SECONDS_IN_ONE_DAY
import pandas as pd
from termcolor import colored

from isitfit.utils import logger


from isitfit.cli.click_descendents import IsitfitCliError


import simple_cache
SECONDS_PER_HOUR = 60*60
class SimpleCacheMan:
  """
  Wrapper around
  https://github.com/barisumog/simple_cache
  """

  def __init__(self, filename, namespace):
    self.filename = filename
    self.namespace = namespace

  def key_with_namespace(self, key):
    k2 = "%s-%s"%(self.namespace, key)
    return k2

  def load_key(self, key):
    k2 = self.key_with_namespace(key)
    return simple_cache.load_key(filename=self.filename, key=k2)

  def save_key(self, key, value):
    k2 = self.key_with_namespace(key)
    simple_cache.save_key(filename=self.filename, key=k2, value=value, ttl=SECONDS_PER_HOUR)


class BaseIterator:
  """
  Iterator design pattern
  Iterates over all CPU performance dataframes
  https://en.wikipedia.org/wiki/Iterator_pattern#Python
  """

  service_name = None
  service_description = None
  paginator_name = None
  paginator_entryJmespath = None
  paginator_exception = None
  entry_keyId = None
  entry_keyCreated = None


  def __init__(self, filter_region=None, tqdmman=None):
    # some defaults
    if tqdmman is None:
      from tqdm import tqdm
      tqdmman = tqdm

    # filter for certain region
    self.filter_region = filter_region

    # list of cluster ID's for which data is not available
    self.rc_noData = []

    # list of regions to skip
    self.region_include = []

    # in case of just_count=True, region_include is ignored since it is not yet populated
    # Set this flag to use region_include, eg if it is loaded from cache or if counting first pass is done
    self.regionInclude_ready = False

    # list to gather AccessDenied errors
    self.region_accessdenied = []

    # flag to display the access denied message only once
    self.displayed_accessdenied = False

    # init cache
    self._initCache()

    # count of entries
    self.n_entry = None

    # handler of local tqdm
    self.tqdmman = tqdmman

    # flag to display "Will skip ... out of ... regions ..." only once
    self.displayed_willskip = False


  def get_regionInclude(self):
    """
    for the sake of mocking in the test tests/cost/redshift/test_cli.py:15
    """
    return self.region_include


  def _initCache(self):
    """
    # try to load region_include from cache
    """
    if self.filter_region is not None:
      self.region_include = [self.filter_region]
      self.regionInclude_ready = True
      return

    # need to use the profile name
    # because a profile could have ec2 in us-east-1
    # whereas another could have ec2 in us-west-1
    import boto3
    profile_name = boto3.session.Session().profile_name

    # cache filename and key to use
    # Update 2019-12-03: move from ~/.isitfit to /tmp/isitfit/
    from isitfit.dotMan import DotMan
    import os
    cache_filename = 'iterator_cache-%s-%s.pkl'%(profile_name, self.service_name)
    cache_filename = os.path.join(DotMan().tempdir(), cache_filename)

    # set of keys to save in local cache file with simple_cache
    self.simpleCacheMan = SimpleCacheMan(filename=cache_filename, namespace="iterator")

    # load cached keys
    ri_cached = self.simpleCacheMan.load_key(key='region_include')
    if ri_cached is not None:
      logger.debug("Loading regions containing EC2 from cache file")
      self.region_include = ri_cached
      self.regionInclude_ready = True

    ri_cached = self.simpleCacheMan.load_key(key='region_accessdenied')
    if ri_cached is not None:
      self.region_accessdenied = ri_cached



  def iterate_core(self, display_tqdm=False):
    fx_l = ['service_name', 'service_description', 'paginator_name', 'paginator_entryJmespath', 'paginator_exception', 'entry_keyId', 'entry_keyCreated']
    for fx_i in fx_l:
      # https://stackoverflow.com/a/9058315/4126114
      if fx_i not in self.__class__.__dict__.keys():
        raise Exception("Derived class should set %s"%fx_i)

    # iterate on regions
    import botocore
    import boto3
    import jmespath
    redshift_regions_full = boto3.Session().get_available_regions(self.service_name)
    import copy
    redshift_regions_sub = copy.deepcopy(redshift_regions_full)
    # redshift_regions_sub = ['us-west-2'] # FIXME

    if self.filter_region is not None:
      if self.filter_region not in redshift_regions_sub:
        msg_err = "Invalid region specified: %s. Supported values: %s"
        msg_err = msg_err%(self.filter_region, ", ".join(redshift_regions_sub))
        raise IsitfitCliError(msg_err, None) # passing None for click context

      # over-ride
      redshift_regions_sub = [self.filter_region]

    # Before iterating, display a message that skipping some regions due to load from cache
    # The following conditions = region_include was loaded from cache
    if self.regionInclude_ready and len(redshift_regions_sub)!=len(self.region_include) and not self.displayed_willskip:
        msg1 = "%s: Will skip %i out of %i regions which were either empty or inaccessible. To re-check, delete the local cache file %s"
        msg1 = msg1%(self.service_description, len(redshift_regions_sub)-len(self.region_include), len(redshift_regions_sub), self.simpleCacheMan.filename)
        import click
        click.echo(colored(msg1, "yellow"))
        self.displayed_willskip = True

    # iterate
    region_iterator = redshift_regions_sub
    if display_tqdm:
      # add some spaces for aligning the progress bars
      desc = "%s, counting in all regions     "%self.service_description
      desc = "%-50s"%desc
      region_iterator = self.tqdmman(region_iterator, total = len(redshift_regions_sub), desc=desc)


    for region_name in region_iterator:
      if self.regionInclude_ready and self.filter_region is None:
        if region_name not in self.region_include:
          # skip since already failed to use it
          continue

      logger.debug("Region %s"%region_name)
      boto3.setup_default_session(region_name = region_name)

      # boto3 clients
      # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/redshift.html#Redshift.Client.describe_logging_status
      # Update 2019-12-09
      #   Unfolding the iterator can cause a rate limiting error for accounts with more than 200 EC2
      #   as reported by u/moofishies on 2019-11-12
      #   Similar to: https://github.com/boto/botocore/pull/891#issuecomment-303526763
      #   The max_attempts config here is increased from the default 4 to decrease the rate limiting chances
      #   https://github.com/boto/botocore/pull/1260
      #   Note that with each extra retry, an exponential backoff is already implemented inside botocore
      #   More: https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
      from botocore.config import Config
      service_client = boto3.client(self.service_name, config=Config(retries={'max_attempts': 10}))

      # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudwatch.html#metric
      self.cloudwatch_resource = boto3.resource('cloudwatch')

      # iterate on service resources, eg ec2 instances, redshift clusters
      paginator = service_client.get_paginator(self.paginator_name)
      rc_iterator = paginator.paginate()
      try:
        region_anyClusterFound = False
        for rc_describe_page in rc_iterator:
          rc_describe_entries = jmespath.search(self.paginator_entryJmespath, rc_describe_page)
          for rc_describe_entry in rc_describe_entries:
            region_anyClusterFound = True
            # add field for region
            rc_describe_entry['Region'] = region_name
            # yield
            yield rc_describe_entry

        if not self.regionInclude_ready and self.filter_region is None:
          if region_anyClusterFound:
            # only include if found clusters in this region
            self.region_include.append(region_name)

      except botocore.exceptions.ClientError as e:
        # Exception that means "no access to region"
        if e.response['Error']['Code']==self.paginator_exception:
          continue

        # eg if user doesnt have access arn:aws:redshift:ap-northeast-1:974668457921:cluster:*
        # it could be because of specific access to region, or general access to the full redshift service
        # Note: capturing this exception means that the region is no longer included in the iterator, but it will still iterate over other regions
        if e.response['Error']['Code']=='AccessDenied':
          self.region_accessdenied.append(e)
          continue

        # Handle error:
        # botocore.exceptions.ClientError: An error occurred (InvalidClientTokenId) when calling the AssumeRole operation: The security token included in the request is invalid.
        # Not sure what this means, but maybe that a role is not allowed to assume into a region?
        # This error can be raised for example with using my local AWS profile "afc_external_readCur".
        # Here is an excerpt from my ~/.aws/credentials file
        # # Role created in Autofitcloud giving access to shadiakiki1986 to read CUR S3
        # [afc_external_readCur]
        # role_arn = arn:aws:iam::123456789:role/external-read-athena-role-ExternalReadCURRole-abcdef
        # source_profile = a_user_profile_not_a_role
        # region = us-east-1
        if e.response['Error']['Code']=='InvalidClientTokenId':
          continue

        # after setting up the InvalidClientTokenId filter above on the profile afc_external_readCur,
        # faced error: botocore.exceptions.ClientError: An error occurred (UnauthorizedOperation) when calling the DescribeInstances operation: You are not authorized to perform this operation.
        if e.response['Error']['Code']=='UnauthorizedOperation':
          continue

        # all other exceptions raised
        raise e

    # before exiting, check if a count just completed, and mark region_include as usable
    if not self.regionInclude_ready and self.filter_region is None:
      self.regionInclude_ready = True

      # save to cache
      self.simpleCacheMan.save_key(key='region_include', value=self.region_include)
      self.simpleCacheMan.save_key(key='region_accessdenied', value=self.region_accessdenied)


    # before exiting, if got some AccessDenied errors, display to user
    # Note 1: originally, I wanted to break the iterator on the 1st AccessDenied error,
    # thinking that it's because the user doesn't have permission to the service as a whole.
    # Later, I figured out that maybe the user has permission to a subset of regions,
    # in which case getting an error on region R1 is normal,
    # and the iterator should still proceed to the next region R2.
    if not self.displayed_accessdenied and len(self.region_accessdenied)>0:
      # 1st part goes to stdout
      msgx = "AWS returned AccessDenied errors on %i out of %i regions. Use `isitfit --verbose ...` and re-run the command for more details"
      msgx = msgx%(len(self.region_accessdenied), len(redshift_regions_sub))
      import click
      click.echo(colored(msgx, "yellow"))

      # 2nd part is too long, send it to --verbose
      msg2 = "\n".join(["- %s"%str(e) for e in self.region_accessdenied])
      msgx = "Here are the full error messages:\n%s"
      msgx = msgx%(msg2)
      logger.info(colored(msgx, "yellow"))

      self.displayed_accessdenied = True


  def count(self):
      # method 1
      # ec2_it = self.ec2_resource.instances.all()
      # return len(list(ec2_it))

    if self.n_entry is not None:
      return self.n_entry

    self.n_entry = len(list(self.iterate_core(True)))

    # interim result for timer data to calculate performance (seconds per ec2 or seconds per rds)
    from isitfit.utils import ping_matomo
    ping_matomo("/cost/base_iterator/BaseIterator/count?service=%s&n_entry=%s&n_region=%s"%(self.service_name, self.n_entry, len(self.region_include)))

    # send message to logs for info
    if self.n_entry==0 and len(self.region_include)==0:
      msg_count = "Found no %s"
      logger.info(msg_count%(self.service_description))
    else:
      msg_count = "Found a total of %i %s in %i region(s) (other regions do not hold any %s)"
      logger.info(msg_count%(self.n_entry, self.service_description, len(self.region_include), self.service_name))

    return self.n_entry


  def __iter__(self):
    for rc_describe_entry in self.iterate_core(False):
        #print("response, entry")
        #print(rc_describe_entry)

        # if not available yet (eg creating), still include analysis in case of past data
        #if rc_describe_entry['ClusterStatus'] != 'available':
        #    self.rc_noData.append(rc_id)
        #    continue

        if self.entry_keyId not in rc_describe_entry:
          # no ID, weird
          continue

        rc_id = rc_describe_entry[self.entry_keyId]

        if self.entry_keyCreated not in rc_describe_entry:
          # no creation time yet, maybe in process
          self.rc_noData.append(rc_id)
          continue

        rc_created = rc_describe_entry[self.entry_keyCreated]

        # None below is a placeholder for ec2_obj in case of ec2
        yield rc_describe_entry, rc_id, rc_created, None







from isitfit.utils import logger
import pandas as pd
import redis


from ..utils import SECONDS_IN_ONE_DAY
SECONDS_IN_10MINS = 60*10

class RedisPandas:
  """
  Python class that manages caching pandas dataframes to redis
  https://stackoverflow.com/a/57986261/4126114
  """
  def __init__(self):
    self.redis_args = {}
    self.redis_client = None
    self.pyarrow_context = None

  def fetch_envvars(self):
    # check redis parameters if set for caching
    import os
    for k1, k2 in [
        ('host', "ISITFIT_REDIS_HOST"),
        ('port', "ISITFIT_REDIS_PORT"),
        ('db', "ISITFIT_REDIS_DB")
      ]:
      self.redis_args[k1] = os.getenv(k2, None)

  def isSetup(self):
    return all(self.redis_args.values())

  def connect(self):
    logger.info("Connecting to redis cache")
    logger.debug(self.redis_args)
    import pyarrow as pa

    self.redis_client = redis.Redis(**self.redis_args)
    self.pyarrow_context = pa.default_serialization_context()

  def isReady(self):
    return self.redis_client is not None

  def set(self, key, df):
    # Note that in case data was not found, eg in mainManager._cloudwatch_metrics_core, an empty dataframe is returned (and thus passed in here)
    pybytes = self.pyarrow_context.serialize(df).to_buffer().to_pybytes()

    # if dataframe with shape[0]==0, raise (no longer supported)
    # Update 2019-12-17 Actually Cloudtrail needs to store an empty dataframe if there are no events in the last 90 days
    if not callable(df):
      from isitfit.cli.click_descendents import IsitfitCliError

      if type(df) != pd.DataFrame:
        raise IsitfitCliError("Internal dev error: Only caching of callables or pandas dataframes supported as of isitfit 0.19")

      # Check comment above about Cloudtrail needing to store empty dataframes
      #if df.shape[0]==0:
      #  raise IsitfitCliError("Internal dev error: caching empty dataframes is no longer supported as of isitfit 0.19")

    # set expiration of key-value pair to be 1 day if data was found, 10 minutes otherwise
    ex = SECONDS_IN_10MINS if callable(df) else SECONDS_IN_ONE_DAY
    # https://redis-py.readthedocs.io/en/latest/#redis.Redis.set
    self.redis_client.set(name=key, value=pybytes, ex=ex)

  def get(self, key):
    try:
      v1 = self.redis_client.get(key)
    except redis.exceptions.ResponseError as e:
      msg = 'Redis error: {e.__class__.__module__}.{e.__class__.__name__}: {e}'.format(e=e)
      # eg, 'redis.exceptions.ResponseError: invalid DB index'
      if 'invalid DB index' in str(e):
        import os
        msg += "\nHint: Check that environment variable ISITFIT_REDIS_DB=%s is a valid redis database index"%(os.getenv("ISITFIT_REDIS_DB", None))
        msg += "\nFor more info: https://stackoverflow.com/questions/13386053/how-do-i-change-between-redis-database"

      from isitfit.cli.click_descendents import IsitfitCliError
      raise IsitfitCliError(msg)

    if not v1: return v1
    v2 = self.pyarrow_context.deserialize(v1)
    return v2

  def handle_pre(self, context_pre):
        # set up caching if requested
        self.fetch_envvars()
        if self.isSetup():
          self.connect()

        # 0th pass to count
        n_ec2_total = context_pre['n_ec2_total']

        # if more than 10 servers, recommend caching with redis
        cond_prompt = n_ec2_total > 10 and not self.isSetup()
        if cond_prompt:
            from termcolor import colored
            logger.warning(colored(
"""Since the number of EC2 instances is %i,
it is recommended to use redis for caching of downloaded CPU/memory metrics.
To do so
- install redis

    [sudo] apt-get install redis-server

- export environment variables

    export ISITFIT_REDIS_HOST=localhost
    export ISITFIT_REDIS_PORT=6379
    export ISITFIT_REDIS_DB=0

where ISITFIT_REDIS_DB is the ID of an unused database in redis.

And finally re-run isitfit as usual.
"""%n_ec2_total, "yellow"))
            import click
            # not using abort=True so that I can send a cusomt message in the abort
            continue_wo_redis = click.confirm(colored('Would you like to continue without redis caching? ', 'cyan'), abort=False, default=True)
            if not continue_wo_redis:
                from isitfit.cli.click_descendents import IsitfitCliError
                raise IsitfitCliError("Aborting to set up redis.", context_pre['click_ctx'])


        # done
        return context_pre


from isitfit.utils import logger, NoCloudwatchException, HostNotFoundInDdg, DataNotFoundForHostInDdg

from isitfit.utils import myreturn
class MetricCacheMixin:
    """
    Mixin for metrics_* classes to get Caching
    """
    def __init__(self, cache_man):
      """
      cache_man - RedisPandasCacheManager
      """
      self.cache_man = cache_man
      super().__init__()


    def get_key(self, host_id):
      raise Exception("Define in derived/mixin")


    def get_metrics_base(self, rc_describe_entry, rc_id, rc_created):
      raise Exception("Define in derived/mixin")


    def get_metrics_derived(self, rc_describe_entry, rc_id, rc_created):
        # check cache first
        cache_key = self.get_key(rc_id)

        if self.cache_man.isReady():
          df_cache = self.cache_man.get(cache_key)
          if df_cache is None:
            # not found
            pass
          else:
            if callable(df_cache):
              # if one of the raise_* functions is cached
              df_cache()

            elif type(df_cache) is pd.DataFrame:
              if df_cache.shape[0]==0:
                # found but no data
                raise Exception("As of isitfit 0.19, empty dataframes are no longer cached")
              else:
                logger.debug("Found metrics (datadog? cloudwatch?)  in redis cache for %s, and data.shape[0] = %i"%(rc_id, df_cache.shape[0]))
                # replace with None if .shape[0]==0
                df_cache = myreturn(df_cache)
                # done
                return df_cache
            else:
              raise ValueError("Invalid value in cache for %s"%cache_key)

        # if no cache, then download
        df_fresh = pd.DataFrame() # use an empty dataframe in order to distinguish when getting from cache if not available in cache or data not found but set in cache
        def error2func(error):
          def raiseme(): raise error
          return raiseme

        try:
          df_fresh = self.get_metrics_base(rc_describe_entry, rc_id, rc_created)
        except HostNotFoundInDdg as error:
          df_fresh = error2func(error)
        except DataNotFoundForHostInDdg as error:
          df_fresh = error2func(error)
        except NoCloudwatchException as error:
          df_fresh = error2func(error)
        except:
          # anything else should bubble up
          raise

        # if caching enabled, store it for later fetching
        # https://stackoverflow.com/a/57986261/4126114
        # Note that this even stores the result if it was "None" (meaning that no data was found)
        if self.cache_man.isReady():
          self.cache_man.set(cache_key, df_fresh)

        # if exception
        if callable(df_fresh):
          df_fresh()

        # done
        return myreturn(df_fresh)




from isitfit.cost.ec2_analyze import BinCapUsed
import datetime as dt
import pytest
import pandas as pd

@pytest.fixture
def FakeMm():
    class FakeMm:
      StartTime = dt.datetime(2019,1,15)
      EndTime = dt.datetime(2019,4,15)

    return FakeMm


class TestBinCapUsedHandlePre:
  def test_preNoBreak(self, FakeMm):
    bcs = BinCapUsed()
    ret = bcs.handle_pre({'mainManager': FakeMm()})
    assert ret is not None

  def test_3m(self, FakeMm):
    bcs = BinCapUsed()
    bcs.handle_pre({'mainManager': FakeMm()})

    e = pd.DataFrame([
        (dt.date(2019,1,31), 0, 0, 0, frozenset([]), dt.date(2019,1,31), dt.date(2019,1,1), ),
        (dt.date(2019,2,28), 0, 0, 0, frozenset([]), dt.date(2019,2,28), dt.date(2019,2,1), ),
        (dt.date(2019,3,31), 0, 0, 0, frozenset([]), dt.date(2019,3,31), dt.date(2019,3,1), ),
        (dt.date(2019,4,30), 0, 0, 0, frozenset([]), dt.date(2019,4,30), dt.date(2019,4,1), ),
      ],
      columns=['Timestamp', 'capacity_usd', 'used_usd', 'count_analyzed', 'regions_set', 'dt_start', 'dt_end']
    )
    for fx in ['Timestamp', 'dt_start', 'dt_end']: e[fx] = pd.to_datetime(e[fx])
    e.set_index('Timestamp', inplace=True)

    pd.testing.assert_frame_equal(e, bcs.df_bins)


class TestBinCapUsedPerEc2:
  def test_1m(self, FakeMm):
    # prepare input
    df1 = pd.DataFrame([
          (dt.date(2019,1,15), 10, 50),
          (dt.date(2019,1,16), 12, 50),
          (dt.date(2019,1,17), 12, 50),
        ],
        columns=['Timestamp','capacity_usd','used_usd']
      )

    # calculate
    bcs = BinCapUsed()
    bcs.handle_pre({'mainManager': FakeMm()})
    ctx = {'ec2_df': df1, 'ec2_dict': {'Region': 'us-west-2'}}
    bcs.per_ec2(ctx)
    bcs.per_ec2(ctx)

    # expected
    e = pd.DataFrame([
        (dt.date(2019,1,31), 68, 300, 2, frozenset(['us-west-2']), dt.date(2019,1,15), dt.date(2019,1,17), ),
        (dt.date(2019,2,28),  0,   0, 0, frozenset([]),            dt.date(2019,2,28), dt.date(2019,2, 1), ),
        (dt.date(2019,3,31),  0,   0, 0, frozenset([]),            dt.date(2019,3,31), dt.date(2019,3, 1), ),
        (dt.date(2019,4,30),  0,   0, 0, frozenset([]),            dt.date(2019,4,30), dt.date(2019,4, 1), ),
      ],
      columns=['Timestamp', 'capacity_usd', 'used_usd', 'count_analyzed', 'regions_set', 'dt_start', 'dt_end']
    )
    for fx in ['Timestamp', 'dt_start', 'dt_end']: e[fx] = pd.to_datetime(e[fx])
    e.set_index('Timestamp', inplace=True)

    # test expected = actual
    pd.testing.assert_frame_equal(e, bcs.df_bins)


  def test_3m(self, FakeMm):
    # prepare input
    s_ts = pd.date_range(start=dt.date(2019,1,15), end=dt.date(2019,4,15), freq='D')

    # parameters for simple case, no fluctuations
    cap1 = 10 # USD/day
    cap2 = 20 # USD/day

    #import numpy as np
    # s_used = np.random.rand(len(s_ts)) # random usage between 0 and 100%
    s_used = 0.3 # 30% usage

    # dataframes
    df1 = pd.DataFrame({
      'Timestamp': s_ts,
      'capacity_usd': cap1,
      'used_usd': s_used*cap1
    })
    df2 = pd.DataFrame({
      'Timestamp': s_ts,
      'capacity_usd': cap2,
      'used_usd': s_used*cap2
    })

    # int for simplicity
    df1['used_usd'] = df1['used_usd'].astype(int)
    df2['used_usd'] = df2['used_usd'].astype(int)

    # calculate
    bcs = BinCapUsed()
    bcs.handle_pre({'mainManager': FakeMm()})
    ctx1 = {'ec2_df': df1, 'ec2_dict': {'Region': 'us-west-2'}}
    bcs.per_ec2(ctx1)
    ctx2 = {'ec2_df': df2, 'ec2_dict': {'Region': 'us-west-2'}}
    bcs.per_ec2(ctx2)

    # expected
    e = pd.DataFrame([
        (dt.date(2019,1,31), 510, 153, 2, frozenset(['us-west-2']), dt.date(2019,1,15), dt.date(2019,1,31), ),
        (dt.date(2019,2,28), 840, 252, 2, frozenset(['us-west-2']), dt.date(2019,2, 1), dt.date(2019,2,28), ),
        (dt.date(2019,3,31), 930, 279, 2, frozenset(['us-west-2']), dt.date(2019,3, 1), dt.date(2019,3,31), ),
        (dt.date(2019,4,30), 450, 135, 2, frozenset(['us-west-2']), dt.date(2019,4, 1), dt.date(2019,4,15), ),
      ],
      columns=['Timestamp', 'capacity_usd', 'used_usd', 'count_analyzed', 'regions_set', 'dt_start', 'dt_end']
    )
    for fx in ['Timestamp', 'dt_start', 'dt_end']: e[fx] = pd.to_datetime(e[fx])
    e.set_index('Timestamp', inplace=True)

    # test expected = actual
    pd.testing.assert_frame_equal(e, bcs.df_bins)


class TestBinCapUsedAfterAll:
  def test_preNoBreak(self, FakeMm):
    bcs = BinCapUsed()
    ret = bcs.handle_pre({'mainManager': FakeMm()})
    assert ret is not None

  def test_3m(self, FakeMm):
    bcs = BinCapUsed()
    bcs.handle_pre({'mainManager': FakeMm()})
    bcs.after_all({})

    import numpy as np
    e = pd.DataFrame([
        (dt.date(2019,1,31), 0, 0, 0, frozenset([]), np.nan, np.nan, 0, '0', ),
        (dt.date(2019,2,28), 0, 0, 0, frozenset([]), np.nan, np.nan, 0, '0', ),
        (dt.date(2019,3,31), 0, 0, 0, frozenset([]), np.nan, np.nan, 0, '0', ),
        (dt.date(2019,4,30), 0, 0, 0, frozenset([]), np.nan, np.nan, 0, '0', ),
      ],
      columns=['Timestamp', 'capacity_usd', 'used_usd', 'count_analyzed', 'regions_set', 'dt_start', 'dt_end', 'used_pct', 'regions_str']
    )
    for fx in ['Timestamp', 'dt_start', 'dt_end']: e[fx] = pd.to_datetime(e[fx])
    e.set_index('Timestamp', inplace=True)

    pd.testing.assert_frame_equal(e, bcs.df_bins)



@pytest.fixture
def example_dataframe():
    import datetime as dt
    d = pd.date_range(dt.date(2019,9,1), dt.date(2019,12,1))
    df = pd.DataFrame({'Timestamp': d, 'a': range(len(d))})
    df['Timestamp'] = pd.to_datetime(df.Timestamp)
    dt_start = df.Timestamp.min()
    dt_end = df.Timestamp.max()
    df = df.set_index('Timestamp')
    return df, dt_start, dt_end


class TestBinCapUsedResample:
  def test_07d(self, example_dataframe):
    dfi, dt_start, dt_end = example_dataframe

    bcs = BinCapUsed()
    bcs._set_freq(7)

    dfe_actual = bcs.do_resample_end(dfi).sum()
    dfs_actual = bcs.do_resample_start(dfi).sum()
    #dfe_actual = bcs.fix_resample_end(dfe_actual, dfs_actual)
    #dfs_actual = bcs.fix_resample_start(dfs_actual, dfe_actual, dt_start, dt_end)

    idxs_expected = pd.date_range(dt.date(2019,9,1), dt.date(2019,12,1))
    assert (dfs_actual.index == idxs_expected).all()

    # notice this is different than the "df_daily.resample('1D', label='right')" below
    idxe_expected = pd.date_range(dt.date(2019,9,1), dt.date(2019,12,1))
    assert (dfe_actual.index == idxe_expected).all()


  def test_30d(self, example_dataframe):
    dfi, dt_start, dt_end = example_dataframe

    bcs = BinCapUsed()
    bcs._set_freq(30)

    dfe_actual = bcs.do_resample_end(dfi).sum()
    dfs_actual = bcs.do_resample_start(dfi).sum()
    #dfe_actual = bcs.fix_resample_end(dfe_actual, dfs_actual)
    #dfs_actual = bcs.fix_resample_start(dfs_actual, dfe_actual, dt_start, dt_end)

    idxs_expected = [
      dt.date(2019, 8,26),
      dt.date(2019, 9, 2), dt.date(2019, 9, 9), dt.date(2019, 9,16), dt.date(2019, 9,23),
      dt.date(2019, 9,30),
      dt.date(2019,10, 7), dt.date(2019,10,14), dt.date(2019,10,21), dt.date(2019,10,28),
      dt.date(2019,11, 4), dt.date(2019,11,11), dt.date(2019,11,18), dt.date(2019,11,25),
    ]
    assert (dfs_actual.index == idxs_expected).all()

    idxe_expected = [
      dt.date(2019, 9, 1), dt.date(2019, 9, 8), dt.date(2019, 9,15), dt.date(2019, 9,22),
      dt.date(2019, 9,29),
      dt.date(2019,10, 6), dt.date(2019,10,13), dt.date(2019,10,20), dt.date(2019,10,27),
      dt.date(2019,11, 3), dt.date(2019,11,10), dt.date(2019,11,17), dt.date(2019,11,24),
      dt.date(2019,12, 1),
    ]
    assert (dfe_actual.index == idxe_expected).all()


  def test_60d(self, example_dataframe):
    dfi, dt_start, dt_end = example_dataframe

    bcs = BinCapUsed()
    bcs._set_freq(60)

    dfe_actual = bcs.do_resample_end(dfi).sum()
    dfs_actual = bcs.do_resample_start(dfi).sum()
    #dfe_actual = bcs.fix_resample_end(dfe_actual, dfs_actual)
    #dfs_actual = bcs.fix_resample_start(dfs_actual, dfe_actual, dt_start, dt_end)

    idxs_expected = [
      dt.date(2019, 9, 1), dt.date(2019, 9,15),
      dt.date(2019,10, 1), dt.date(2019,10,15),
      dt.date(2019,11, 1), dt.date(2019,11,15),
      dt.date(2019,12, 1),
    ]
    assert (dfs_actual.index == idxs_expected).all()

    # notice this is different than idx_exp_1SM_right_right below
    # Update 2019-12-17 without the fix_resample_* function, this is the same as idx_exp_1SM_right_right
    #idxe_expected = [
    #  dt.date(2019, 9,14), dt.date(2019, 9,30),
    #  dt.date(2019,10,14), dt.date(2019,10,31),
    #  dt.date(2019,11,14), dt.date(2019,11,30),
    #  dt.date(2019,12,14),
    #]
    idxe_expected = [
      dt.date(2019, 9,15), dt.date(2019, 9,30),
      dt.date(2019,10,15), dt.date(2019,10,31),
      dt.date(2019,11,15), dt.date(2019,11,30),
      dt.date(2019,12,15),
    ]
    assert (dfe_actual.index == idxe_expected).all()


  def test_90d(self, example_dataframe):
    dfi, dt_start, dt_end = example_dataframe

    bcs = BinCapUsed()
    bcs._set_freq(90)

    dfe_actual = bcs.do_resample_end(dfi).sum()
    dfs_actual = bcs.do_resample_start(dfi).sum()
    #dfe_actual = bcs.fix_resample_end(dfe_actual, dfs_actual)
    #dfs_actual = bcs.fix_resample_start(dfs_actual, dfe_actual, dt_start, dt_end)

    idxs_expected = [
      dt.date(2019, 9, 1),
      dt.date(2019,10, 1),
      dt.date(2019,11, 1),
      dt.date(2019,12, 1),
    ]
    assert (dfs_actual.index == idxs_expected).all()

    idxe_expected = [
      dt.date(2019, 9,30),
      dt.date(2019,10,31),
      dt.date(2019,11,30),
      dt.date(2019,12,31),
    ]
    assert (dfe_actual.index == idxe_expected).all()



class TestPandasResample:
  """
  General tests on pandas resample method
  """
  def test_resample(self, example_dataframe):
    """
    Setting closed=... only affects 1 case (check comments below)
    """
    df_daily, dt_start, dt_end = example_dataframe

    # expected indeces
    idx_exp_1WMON_left_left = [
      dt.date(2019, 8,26),
      dt.date(2019, 9, 2), dt.date(2019, 9, 9), dt.date(2019, 9,16), dt.date(2019, 9,23),
      dt.date(2019, 9,30),
      dt.date(2019,10, 7), dt.date(2019,10,14), dt.date(2019,10,21), dt.date(2019,10,28),
      dt.date(2019,11, 4), dt.date(2019,11,11), dt.date(2019,11,18), dt.date(2019,11,25),
    ]
    col_exp_1WMON_left_left = [
      0,
      28, 77, 126, 175,
      224,
      273, 322, 371, 420,
      469, 518, 567, 616,
    ]
    df_exp_1WMON_left_left = pd.DataFrame({'a': col_exp_1WMON_left_left, 'Timestamp': idx_exp_1WMON_left_left})
    df_exp_1WMON_left_left['Timestamp'] = pd.to_datetime(df_exp_1WMON_left_left.Timestamp)
    df_exp_1WMON_left_left.set_index('Timestamp', inplace=True)

    col_exp_1WMON_left_skip = [
      1,
      35, 84, 133, 182,
      231,
      280, 329, 378, 427,
      476, 525, 574, 531,
    ]
    df_exp_1WMON_left_skip = pd.DataFrame({'a': col_exp_1WMON_left_skip, 'Timestamp': idx_exp_1WMON_left_left})
    df_exp_1WMON_left_skip['Timestamp'] = pd.to_datetime(df_exp_1WMON_left_skip.Timestamp)
    df_exp_1WMON_left_skip.set_index('Timestamp', inplace=True)


    idx_exp_1WSUN_right_right = [
      dt.date(2019, 9, 1), dt.date(2019, 9, 8), dt.date(2019, 9,15), dt.date(2019, 9,22),
      dt.date(2019, 9,29),
      dt.date(2019,10, 6), dt.date(2019,10,13), dt.date(2019,10,20), dt.date(2019,10,27),
      dt.date(2019,11, 3), dt.date(2019,11,10), dt.date(2019,11,17), dt.date(2019,11,24),
      dt.date(2019,12, 1),
    ]

    idx_exp_1SMS_left_left = [
      dt.date(2019, 9, 1), dt.date(2019, 9,15),
      dt.date(2019,10, 1), dt.date(2019,10,15),
      dt.date(2019,11, 1), dt.date(2019,11,15),
      dt.date(2019,12, 1),
    ]

    # notice this is different than test_60d above
    idx_exp_1SM_right_right = [
      dt.date(2019, 9,15), dt.date(2019, 9,30),
      dt.date(2019,10,15), dt.date(2019,10,31),
      dt.date(2019,11,15), dt.date(2019,11,30),
      dt.date(2019,12,15),
    ]

    idx_exp_1MS_left_left = [
      dt.date(2019, 9, 1),
      dt.date(2019,10, 1),
      dt.date(2019,11, 1),
      dt.date(2019,12, 1),
    ]

    idx_exp_1M_right_right = [
      dt.date(2019, 9,30),
      dt.date(2019,10,31),
      dt.date(2019,11,30),
      dt.date(2019,12,31),
    ]


    # 1D should be the same
    df_resampled = df_daily.resample('1D', label='left', closed='left').sum()
    pd.testing.assert_frame_equal(df_daily, df_resampled)
    assert (df_daily.index == df_resampled.index).all()

    # again but skip closed
    df_resampled = df_daily.resample('1D', label='left').sum()
    pd.testing.assert_frame_equal(df_daily, df_resampled)
    assert (df_daily.index == df_resampled.index).all()

    # again with right
    df_resampled = df_daily.resample('1D', label='right', closed='right').sum()
    pd.testing.assert_frame_equal(df_daily, df_resampled)
    assert (df_daily.index == df_resampled.index).all()

    # again with right, but without setting closed
    df_resampled = df_daily.resample('1D', label='right').sum()
    # pd.testing.assert_frame_equal(df_daily, df_resampled)
    assert not (df_daily.index == df_resampled.index).all() # <<<<<<<<<<<< notice expected difference

    # 1W-MON
    df_resampled = df_daily.resample('1W-MON', label='left', closed='left').sum()
    assert (idx_exp_1WMON_left_left == df_resampled.index).all()
    pd.testing.assert_frame_equal(df_exp_1WMON_left_left, df_resampled)

    # again with skip closed
    df_resampled = df_daily.resample('1W-MON', label='left').sum()
    assert (idx_exp_1WMON_left_left == df_resampled.index).all()
    pd.testing.assert_frame_equal(df_exp_1WMON_left_skip, df_resampled) # <<<<<<<<<<<<< notice eventhough the index is the same, the summed values are different

    # 1W-SUN
    df_resampled = df_daily.resample('1W-SUN', label='right', closed='right').sum()
    assert (idx_exp_1WSUN_right_right == df_resampled.index).all()

    # again with skip closed
    df_resampled = df_daily.resample('1W-SUN', label='right').sum()
    assert (idx_exp_1WSUN_right_right == df_resampled.index).all()

    # 1SMS
    df_resampled = df_daily.resample('1SMS', label='left', closed='left').sum()
    assert (idx_exp_1SMS_left_left == df_resampled.index).all()

    # again with skip closed
    df_resampled = df_daily.resample('1SMS', label='left').sum()
    assert (idx_exp_1SMS_left_left == df_resampled.index).all()

    # 1SM
    df_resampled = df_daily.resample('1SM', label='right', closed='right').sum()
    assert (idx_exp_1SM_right_right == df_resampled.index).all()

    # again with skip closed
    df_resampled = df_daily.resample('1SM', label='right').sum()
    assert (idx_exp_1SM_right_right == df_resampled.index).all()

    # 1MS
    df_resampled = df_daily.resample('1MS', label='left', closed='left').sum()
    assert (idx_exp_1MS_left_left == df_resampled.index).all()

    # again with skip closed
    df_resampled = df_daily.resample('1MS', label='left').sum()
    assert (idx_exp_1MS_left_left == df_resampled.index).all()

    # 1M
    df_resampled = df_daily.resample('1M', label='right', closed='right').sum()
    assert (idx_exp_1M_right_right == df_resampled.index).all()

    # again with skip closed
    df_resampled = df_daily.resample('1M', label='right').sum()
    assert (idx_exp_1M_right_right == df_resampled.index).all()

